{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name drcompress1}
{@description Dynamic range compression using non-linear high-pass filter and regularization method 1}

{$I directives.pas}
uses simage,getparam,sysutils;
var par         : tparams;
    cc          : CSImg_Config;
    img         : TSImg_FPImage;
    mask        : TSImg_RDImage;
    ifn,ofn,mfn : ansistring;
    l,m,s,r,c   : double;
    i,n         : longint;
    ss          : array of double;
    mi          : CSImg_MetaInfo;

begin
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        ifn:='';        defparam('<infile>','Source image (default: stdin)',ifn);
        ofn:='';        defparam('<outfile>','Destination image (default: stdout)',ofn);
        mfn:='';        defparam('-mask','Mask image calculated by getmask', mfn);
        l:=0.05;        defparam('-l','Black level threshold as quantile or <0 or disable it.. Default: 5%',l);
        m:=0.75;        defparam('-m','25% compression threshold as quantile. Default: 75%',m);
        s:=3;           defparam('-s','Minimum detail size to block (default: 3)',s);
        n:=3;           defparam('-n','Number of filters (default: 3)',n);
        r:=3;           defparam('-sr','Size factor for additional filters (relevant if n>1, default: 3)',r);
        c:=1;           defparam('-c','Saturation preservation. Valid: 0..1. Default: 1',c);
                        defparam('-o','Destination image',ofn); deffmt('-o','<outfile>');
                        defparam('-h','Help',helpf);
        run;
        done;
    end;

    if n<1 then n:=1;
    setlength(ss, n);
    for i:=0 to n-1 do
    begin
        ss[i]:=s;
        s*=r;
    end;

    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug('drcompress1: values: source='+ifn+' dest='+ofn+' mask='+mfn+' l='+real2str(l)+' m='+real2str(m)+' s='+real2str(s)+' n='+int2str(n)+' r='+real2str(r)+' c='+real2str(c));

    img.pt:=nil;
    mi:=CSImg_MetaInfo.create();
    cc.Info('drcompress1: Reading input image '+ifn);
    SImg_ReadImageFile(cc,ifn,img,mi);
    mi.deleteRaw;

    mask.pt:=nil;
    if mfn<>'' then
    begin
        cc.Info('drcompress1: Reading mask image '+mfn);
        SImg_ReadImageFile(cc,mfn,mask);
    end;

    cc.Info('drcompress1: High-pass filtering using '+int2str(cc.threads)+' threads');
    SImg_DRCompress1(cc, img,mask, l,m, ss, c);

    cc.Info('drcompress1: Writing output image '+ofn);
    SImg_WriteImageFile(cc,ofn,img,mi);

    SImg_Free(cc,img);
    if mask.pt<>nil then SImg_Free(cc,mask);
    mi.destroy;
    cc.destroy;
end.
