{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name drcompress3}
{@description Dynamic range compression using non-linear high-pass filter and regularization method 3}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                 : tparams;
    cc                  : CSImg_Config;
    img                 : TSImg_FPImage;
    mask                : TSImg_RDImage;
    ifn,ofn,mfn         : ansistring;
    l,smin,smax,ravg    : double;
    c,reg,sreg,sig,s    : double;
    mi                  : CSImg_MetaInfo;

begin
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        ifn:='';        defparam('<infile>','Source image (default: stdin)',ifn);
        ofn:='';        defparam('<outfile>','Destination image (default: stdout)',ofn);
        mfn:='';        defparam('-mask','Mask image calculated by getmask', mfn);
        l:=0.05;        defparam('-l','Black level threshold as quantile or <0 or disable it. Default: 5%',l);
        smin:=1.5;      defparam('-sa','Smallest detail filter as radius. Should be larger than super-resolution factor. Default: 1.5.',smin);
        smax:=-1;       defparam('-sz','Largest filter as radius. Set to 50% of the smaller image dimension by default', smax);
        ravg:=0.4;      defparam('-r','Average dynamic range compression. Default: 0.4', ravg);
        reg:=0.5;       defparam('-n','Compression regularization relative to noise. Default: 0.5 (less compression/noise amplification for darker zones)', reg);
        sreg:=0;        defparam('-ns','Used median filter instead of minimum filter below that size. Default: 0', sreg);
        sig:=-1;        defparam('-sig','Noise floor. <=0 for auto-detect. Default: -1',sig);
        c:=1;           defparam('-c','Saturation preservation. Valid: 0..1. Default: 1',c);
                        defparam('-o','Destination image',ofn); deffmt('-o','<outfile>');
                        defparam('-h','Help',helpf);
        run;
//      if not vardef(@smed) then smed:=smin*1.4;
        done;
    end;

    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug(cc.name+': values: source='+ifn+' dest='+ofn+' mask='+mfn+' l='+real2str(l)+' smin='+real2str(smin)+' smax='+real2str(smax)+' ravg='+real2str(ravg,-1)+' reg='+real2str(reg)+' sreg='+real2str(sreg)+' c='+real2str(c)+' sig='+real2str(sig));

    img.pt:=nil;
    mi:=CSImg_MetaInfo.create();
    cc.Info(cc.name+': Reading input image '+ifn);
    SImg_ReadImageFile(cc,ifn,img,mi);
    mi.deleteRaw;

    mask.pt:=nil;
    if mfn<>'' then
    begin
        cc.Info(cc.name+': Reading mask image '+mfn);
        SImg_ReadImageFile(cc,mfn,mask);
    end;

    cc.Info(cc.name+': High-pass filtering using '+int2str(cc.threads)+' threads');
    if (sig<simg_mina) and mi.get('NOISEFLR',s) then sig:=s;

    SImg_DRCompress3(cc, img,mask, l, smin,sreg,smax, ravg, reg,sig, c);
    mi.add('NOISEFLR', real2str(sig,10),'Noise floor');

    cc.Info(cc.name+': Writing output image '+ofn);
    SImg_WriteImageFile(cc,ofn,img,mi);

    SImg_Free(cc,img);
    if mask.pt<>nil then SImg_Free(cc,mask);
    mi.destroy;
    cc.destroy;
end.
