{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name histcutpow}
{@description Crop and scale brightness with regularized non-linear tonal curve }


{$I directives.pas}
uses simage,getparam,sysutils;
var par                 : tparams;
    cc                  : CSImg_Config;
    img                 : TSImg_FPImage;
    fi,fo               : string;
    l,u,b,r,e,s,g,br    : double;
    mi                  : CSImg_MetaInfo;

begin
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        fi:='';     defparam('<infile>','Source image (default: stdin)',fi);
        fo:='';     defparam('<outfile>','Destination image (default: stdout)',fo);
                    defparam('-h','Help',helpf);
        l:=0.10;    defparam('-l','Percentage of the lower quantile (default: 10%)',l);
        u:=0.99999; defparam('-u','Percentage of the lower quantile  (default: 99.999%)',u);
        r:=1;       defparam('-r','Regularization relative to noise floor. Default: 1. Valid: >0 ', r);
        g:=1;       defparam('-g','Power transform exponent. Default: 1. Valid: >0 ', g);
        e:=1;       defparam('-s','Reciprocal luminance exponent. Default: 1. Valid: >0.', e);
        b:=2;       defparam('-b','Brightness correction. Valid (0..1)', b);
        br:=1;      defparam('-br','Region of interest used for brightness adjustment. Default: 1. Recommended: (0.1..1)', br);
                    defparam('-o','Destination image',fo); deffmt('-o','<outfile>');
        run;

        cc:=CSImg_Config.create(SImg_Config_Default);
        cc.debug(cc.name+': values: source='+fi+' dest='+fo+' l='+real2str(l)+' u='+real2str(u)+' r='+real2str(r)+' e='+real2str(e)+' b='+real2str(b));

        mi:=CSImg_MetaInfo.create();
        cc.Info(cc.name+': Reading input image '+fi);
        SImg_ReadImageFile(cc,fi,img,mi);
        mi.deleteRaw;
        if r<1e-3 then r:=0
        else begin
            if not mi.get('NOISEFLR',s) then
            begin
                s:=SImg_NoiseFloor(cc,img,2.5);
                mi.add('NOISEFLR', real2str(s,10), 'Noise floor');
            end;
            r*=s;
        end;

        cc.Info(cc.name+': Histogram transformation');
        b:=sqr(b);
        if e<1e-6 then e:=1e-6;
        SImg_HistCutPow(cc, img, l,u, r,g,1/e, 1-b*br, b);

        cc.Info(cc.name+': Write output image '+fo);
        SImg_WriteImageFile(cc,fo,img,mi);
        SImg_Free(cc,img);

        done;
    end;

    cc.destroy;
end.
