{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name maskor}
{@description Combines two 9 bit masks masks: <output> = <input1> or (<inout2> and p).}
{@longdescription Reads a 8 bit image and applies mask operations}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                 : TParams;
    cc                  : CSImg_Config;
    ri1,ri2             : TSImg_RDImage;
    mi                  : CSImg_MetaInfo;
    ifn1,ifn2,ofn       : ansistring;
    i,p                 : longint;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='<infile1> [<infile2> [<outfile>]]';
        ifn1:=''; defparam('<infile1>','Input mask 1',@ifn1,tpastr,rapdreq);
        ifn2:=''; defparam('<infile2>','Input mask 1 (default: stdin)',ifn2);
        ofn:='';  defparam('<outfile>','Output mask(default: stdout)',ofn);
        p:=255;   defparam('-p','Bit pattern p. Default: 255. Result is <output> = <input1> or (<inout2> and p).',p);
                  defparam('-o','Destination image',ofn); deffmt('-o','<outfile>');
                  defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc.debug(cc.name+': input1=' + ifn1 + ' input2=' + ifn2 + ' output=' + ofn+ ' p=' + int2str(p));

    { input }
    cc.info(cc.name+': Reading input mask 1 '+ifn1);
    mi:=CSImg_MetaInfo.create;
    ri1.pt:=nil;
    SImg_ReadImageFile(cc,ifn1,ri1,mi);
    if ri1.fmt<>SImg_RawByte then cc.error(cc.name+': 8 bit image expected');

    cc.info(cc.name+': Reading input mask 2 '+ifn1);
    ri2.pt:=nil;
    SImg_ReadImageFile(cc,ifn2,ri2,mi);
    if ri2.fmt<>SImg_RawByte then cc.error(cc.name+': 8 bit image expected');
    if ri2.size<>ri1.size then cc.error(cc.name+': Incompatible dimensions of input masks');

    { output }
    cc.info(cc.name+': Generating and writing output image '+ofn);
    for i:=0 to ri1.size-1 do
        ri1.ptb[i]:=ri1.ptb[i] or (ri2.ptb[i] and p);
    SImg_Free(cc,ri2);
    SImg_WriteImageFile(cc,ofn,ri1,mi);
    SImg_Free(cc,ri1);
    mi.destroy;
    cc.destroy;
end.
