{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name sdbcreate}
{@description Creates a stack database with quadratic pixels}
{@londescription For highest resolution and no information loss scale factors less than 0.667 are recommended for
normal images and scale factors less than 1.0 are recommended for Bayer pattern images.}

{$I directives.pas}
uses simage,getparam,sysutils;
var par       : tparams;
    cc        : CSImg_Config;
    ofn       : ansistring;
    nx,ny,f   : longint;
    xc,yc     : double;
    scale,rot : double;
    flip      : boolean;
    fi        : TSImg_FPImage;
    si        : TSImg_SDBInfo;
    mi        : CSImg_MetaInfo;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='[<outfile>]';
        ofn:='';      defparam('<outfile>','Output stack database (default: stdout)',ofn);
                      defparam('-h','Help',helpf);
        nx:=1000;     defparam('-nx','Size in X direction (default: 1000)',nx);
        ny:=1000;     defparam('-ny','Size in Y direction (default: 1000)',ny);
        scale:=1;     defparam('-s','Scale (default: 1)',scale);
        xc:=0.5;      defparam('-xc','X center relative to size (default: 0.5)',xc);
        yc:=0.5;      defparam('-yc','Y center relative to size (default: 0.5)',yc);
        rot:=0;       defparam('-r','Rotate image clockwise by that amount in degree (default: 0)',rot);
        flip:=false;  defparam('-fy','Flip image in Y direction. (rotated first, then flipped)',flip);
        f:=1;         defparam('-f','Number of FWHM ranges (default: 1)',f);
        run;

        done;
    end;
    if f<1 then f:=1;
    cc.debug('sdbcreate: out='+ofn+'  nx='+int2str(nx)+'  ny='+int2str(ny)+' s='+real2str(scale,3)+' xc='+real2str(xc,3)+' yc='+real2str(xc,3)+' rot='+real2str(rot)+' flip='+boolean2str(flip)+' f='+int2str(f));

    //if s>1.0001 then cc.warning('sdbcreate: Scale factors greater than 1 can cause information loss');

    cc.Info('sdbcreate: Creating and writing stack database '+ofn);
    si.x0:=xc*(nx-1);
    si.y0:=yc*(ny-1);
    SImg_StackDB_InitInfo(si, scale,rot,flip);

    mi:=CSImg_MetaInfo.create();
    SImg_StackDB_WriteInfo(si,mi);
    SImg_New(cc,fi, nx,ny,2*f);

    SImg_WriteImageFile(cc,ofn,fi,mi);

    SImg_Free(cc,fi);

    cc.destroy;
end.
