{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

// define[approx_debug]
// define[bgapprox3_debug]

{*************************************************************************}
{*************************************************************************}
{******* internal methodes ***********************************************}
{*************************************************************************}
{*************************************************************************}




{*************************************************************************}
{******* SImg_ApproxW_int ************************************************}
{*************************************************************************}
const SImg_approxw_int_pnmax = 16;
{ calculate scalar products and vector
    fi        : input image
    fw        : weight image
    fp        : polynomial image. If defined, multiplied with polynomials, and DOF are increased by pn
    fb        : approximation image. If defined, DOF is increased by 1
    ya,yz     : region for this thread
    pn        : number of polynomials
    sp        : scalar products (results) for vector and matrix: nc*(pn2 + pn2*(pn2+1)/2)
}
threadprocedure SImg_ApproxW_int_thread1(fi:TSImg_FPImage; fw:TSImg_FPImage; fp:TSImg_FPImage; fb:TSImg_FPImage; ya:longint; yz:longint; pn:longint; sp:pdouble );
var i,j,k,l,c,pn2  : longint;
    x,y,r,s,rx,ry  : double;
    p              : array[0..SImg_approxw_int_pnmax*2] of double;
    d,e,g,h        : ^simgfp;
    f              : ^double;
begin
    rx:=1/(fi.nx-1);
    ry:=1/(fi.ny-1);
    d:=fi.pt + ya*fi.sl;
    e:=fw.pt + ya*fw.sl;
    pn2:=pn;
    g:=nil;
    if fp.pt<>nil then
    begin
        g:=fp.pt + ya*fp.sl;
        pn2*=2;
    end;
    h:=nil;
    if fb.pt<>nil then
    begin
        h:=fb.pt + ya*fb.sl;
        pn2+=1;
    end;
    for i:=ya to yz do
    begin
        y:=i*ry-0.5;
        for j:=0 to fi.nx-1 do
        begin
            x:=j*rx-0.5;
            p[0]:=1;
            p[1]:=x;
            p[2]:=y;
            p[3]:=x*y;
            r:=sqr(x)+sqr(y);
            s:=1;
            for k:=4 to pn-1 do
            begin
                s*=r;
                p[k]:=s;
            end;
            if g<>nil then
            begin
                s:=g^; g+=1;
                for k:=0 to pn-1 do
                    p[k+pn]:=p[k]*s;
            end;
            if h<>nil then
            begin
                p[pn2-1]:=h^;
                h+=1;
            end;
            f:=sp;
            for c:=0 to fi.nc-1 do
            begin
                r:=d^; d+=1;
                s:=e^; e+=1;
                for k:=0 to pn2-1 do
                begin
                    x:=p[k]*s;
                    f[0]+=x*r;
                    for l:=0 to k do
                        f[l+1]+=x*p[l];
                    f+=k+2;
                end;
            end;
        end;
    end;
end;

{ calculate approximation
    fi        : input if dm=true
    fi        : output image, can be equal to fi
    fp        : polynomial image. If defined, multiplied with polynomials, and DOF are increased by pn
    fb        : approximation image. If defined, DOF is increased by 1
    ya,yz     : region for this thread
    pn        : number of polynomials
    a         : coefficients nc*pn2
    dm        : difference mode
}
threadprocedure SImg_ApproxW_int_thread2(fi:TSImg_FPImage; fo:TSImg_FPImage; fp:TSImg_FPImage; fb:TSImg_FPImage; ya:longint; yz:longint; pn:longint; a:pdouble; dm:boolean);
var i,j,k,c,pn2    : longint;
    x,y,r,s,rx,ry  : double;
    p              : array[0..SImg_approxw_int_pnmax*2] of double;
    d,e,g,h        : ^simgfp;
    f              : ^double;
begin
    rx:=1/(fo.nx-1);
    ry:=1/(fo.ny-1);
    d:=fi.pt + ya*fi.sl;
    e:=fo.pt + ya*fo.sl;
    pn2:=pn;
    g:=nil;
    if fp.pt<>nil then
    begin
        g:=fp.pt + ya*fp.sl;
        pn2*=2;
    end;
    h:=nil;
    if fb.pt<>nil then
    begin
        h:=fb.pt + ya*fb.sl;
        pn2+=1;
    end;
    for i:=ya to yz do
    begin
        y:=i*ry-0.5;
        for j:=0 to fo.nx-1 do
        begin
            x:=j*rx-0.5;
            p[0]:=1;
            p[1]:=x;
            p[2]:=y;
            p[3]:=x*y;
            r:=sqr(x)+sqr(y);
            s:=1;
            for k:=4 to pn-1 do
            begin
                s*=r;
                p[k]:=s;
            end;
            if g<>nil then
            begin
                s:=g^; g+=1;
                for k:=0 to pn-1 do
                    p[k+pn]:=p[k]*s;
            end;
            if h<>nil then
            begin
                p[pn2-1]:=h^;
                h+=1;
            end;
            f:=a;
            for c:=0 to fo.nc-1 do
            begin
                r:=0;
                for k:=0 to pn2-1 do
                    r+=f[k]*p[k];
                f+=pn2;
                if dm then
                begin
                    e^:=d^-r;
                    d+=1;
                end
                else e^:=r;
                e+=1;
            end;
        end;
    end;
end;

{ approximate an image with weight by polynomials
    fi   : input image
    fw   : weight
    fp   : polynomial image. If defined, multiplied with polynomials, and DOF are increased by pn. 1 color image
    fb   : approximation image. If defined, DOF is increased by 1. 1 color image
    fo   : output image
    pn   : number of polynomials (recommended: 1,3,4,5,..)
    dm   : false: approximation replaces input
           true: difference replaces input
    pf   : factor control, if not nil:
            input:  vector of factors for approximation functions (one vector per color), no output if all factors are 0
            output: approximation factors (one vector per color)
   approximation functions:
    p_0(x,y)=1
    p_1(x,y)=x
    p_2(x,y)=y
    p_3(x,y)=x*y
    p_3+i=(x^2+y^2)^i
    p_pn(x,y)=fp[i]
    p_(pn+1)(x,y)=fp[i]*x
    ...
    p_(pn2-1)(x,y)=fb[i]
}
procedure SImg_ApproxW_int(cc:CSImg_Config; fi,fw,fp,fb:TSImg_FPImage; var fo:TSImg_FPImage; pn:longint; dm:boolean; pf:pdouble);
var fj          : TSimg_FPImage;
    ma,vx,vy    : matrix;
    sp,f        : ^double;                       { buffer for <p_i,f> and approximation coefficients }
    i,j,k,tn    : longint;
    ya,yz,pn2   : longint;
    r           : double;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;

    if (pn<1) and (fb.pt=nil) then
    begin
        SImg_IncDebugLevel;
        cc.warning('SImg_ApproxW_int: Number of polynomials less than 1: Nothing to do.');
        exit;
    end;
    if pn>SImg_ApproxW_int_pnmax then pn:=SImg_ApproxW_int_pnmax;
    pn2:=pn;

    if not SImg_Recover(cc,fi) then cc.error('SImg_ApproxW_int: Invalid image: fi');
    if not SImg_Recover(cc,fw) then cc.error('SImg_ApproxW_int: Invalid image: fw');
    if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>fi.nc) then cc.error('SImg_ApproxW_int: Incompatible image size: fw');
    if fp.pt<>nil then
    begin
        if not SImg_Recover(cc,fp) then cc.error('SImg_ApproxW_int: Invalid image: fp');
        if fp.size*fi.nc<>fi.size then cc.error('SImg_ApproxW_int: Incompatible image size: fp');
        pn2*=2;
    end;
    if fb.pt<>nil then
    begin
        if not SImg_Recover(cc,fb) then cc.error('SImg_ApproxW_int: Invalid image: fb');
        if fb.size*fi.nc<>fi.size then cc.error('SImg_ApproxW_int: Incompatible image size: fb');
        pn2+=1;
    end;
    if (not SImg_OverwriteAlert(fi,fo)) and (not SImg_OverwriteAlert(fw,fo)) and (not SImg_OverwriteAlert(fp,fo)) and (not SImg_OverwriteAlert(fb,fo)) then SImg_Free(cc,fo);

    tn:=cc.threads;
    k:=pn2 + pn2*(pn2+1) div 2;
    getmem2(sp, tn*fi.nc*k*sizeof(double));
    for i:=0 to tn*fi.nc*k-1 do
        sp[i]:=0;

    { calculate scalar products }
    ya:=0;
    for j:=0 to tn-1 do
    begin
        yz:=fi.ny*(j+1) div tn;
        dothread SImg_ApproxW_int_thread1(fi,fw,fp,fb, ya,yz-1, pn, sp+j*fi.nc*k);
        ya:=yz;
    end;
    threadwait;
    for i:=0 to fi.nc*k-1 do
        for j:=1 to tn-1 do
            sp[i]+=sp[i+j*fi.nc*k];

    { solve for coefficients }
    am2(ma, zmatrix(pn2,pn2));
    am2(vy, zmatrix(1,pn2));
    vx.pt:=nil;
    f:=sp;
    for i:=0 to fi.nc-1 do
    begin
        for j:=0 to pn2-1 do
        begin
            vy.pt[j]:=f[0];
            for k:=0 to j do
            begin
                ma.pt[j*pn2+k]:=f[k+1];
                ma.pt[k*pn2+j]:=f[k+1];
            end;
            f+=j+2;
        end;
        am(vx,solve(ma,vy));
        for j:=0 to pn2-1 do
            sp[i*pn2+j]:=vx.pt[j];
    end;
    free(ma);
    free(vx);
    free(vy);

    if pf<>nil then
    begin
        r:=0;
        for i:=0 to fi.nc*pn2-1 do
            r+=pf[i];
        if r<simg_mina then exit;
        for i:=0 to fi.nc*pn2-1 do
        begin
            r:=sp[i];
            sp[i]:=r*pf[i];
            pf[i]:=r;
        end;
    end;

    fj:=fo;
    if (not SImg_Recover(cc,fo)) or (fo.nx<>fi.nx) or (fo.ny<>fi.ny) or (fo.nc<>fi.nc) then SImg_New(cc, fj, fi.nx,fi.ny,fi.nc, false);
    ya:=0;
    for j:=0 to tn-1 do
    begin
        yz:=fi.ny*(j+1) div tn;
        dothread SImg_ApproxW_int_thread2(fi,fj,fp,fb, ya,yz-1, pn, sp, dm);
        ya:=yz;
    end;
    threadwait;

    if fj.pt<>fo.pt then SImg_Move(cc,fj,fo);

    freemem2(sp);
end;

{ fo=fi }
procedure SImg_ApproxW_int(cc:CSImg_Config; fi,fw,fp,fb:TSImg_FPImage; pn:longint; dm:boolean; pf:pdouble);
begin
    SImg_ApproxW_int(cc, fi,fw,fp,fb, fi, pn,dm,pf);
end;


{*************************************************************************}
{********* SImg_EqualizeWeight_int ***************************************}
{*************************************************************************}
threadprocedure SImg_EqualizeWeight_thread(ti:longint; tn:longint; fw:TSImg_FPImage; fj:TSImg_FPImage);
var i,ia,iz     : longint;
    r,s         : double;
begin
    ia:=(( ti   *fw.ny) div tn)*fw.sl;
    iz:=(((ti+1)*fw.ny) div tn)*fw.sl - 1;
    for i:=ia to iz do
    begin
        r:=sqrt(fj.pt[i]);
        s:=fw.pt[i];
        if r<s then fj.pt[i]:=1
        else fj.pt[i]:=s/r;
    end;
end;

{ Equalize weight
    fw: Input image
    fo: Output image
}
procedure SImg_EqualizeWeight_int(cc:CSImg_Config; fw:TSImg_FPImage; var fo:TSImg_FPImage);
var fj          : TSImg_FPImage;
    a,b,r       : double;
    i,n         : longint;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fw) then cc.error('SImg_EqualizeWeight_int: Invalid image: fi');
    if not sImg_OverwriteAlert(fw,fo) then SImg_Free(cc,fo);
    n:=min(fw.nx,fw.ny);

    { regularization }
    a:=0; b:=simg_mina;
    for i:=0 to fw.size-1 do
    begin
        r:=abs(fw.pt[i]);
        a+=sqr(r);
        b+=r;
    end;
    a:=sqr(a/b*0.05);

    { filtering }
    fj.pt:=nil;
    SImg_New(cc, fj, fw.nx,fw.ny,fw.nc, false);
    for i:=0 to fw.size-1 do
        fj.pt[i]:=sqr(fw.pt[i])+a;
    SImg_Filter(cc, fj,fj, n*0.01);

    { output }
    for i:=0 to cc.threads-1 do
        dothread SImg_EqualizeWeight_thread(i,cc.threads, fw,fj);
    threadwait;

    SImg_Move(cc, fj,fo);
end;



{*************************************************************************}
{*************************************************************************}
{******* global methodes *************************************************}
{*************************************************************************}
{*************************************************************************}




{*************************************************************************}
{******* SImg_BgApprox2 **************************************************}
{*************************************************************************}
{ constant background approximation
    res    : result buffer for at least fi.nc values
}
procedure SImg_BgApprox2(cc:CSImg_Config; fi:TSImg_FPImage; res:psimgfp);
var  fj           : TSImg_FPImage;
     d1,d2,d3,d4  : ^simgfp;
     x,y,c        : longint;
     r            : simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BgApprox2 fi='+SImg_InfoStr2(fi));
    if not SImg_Recover(cc,fi) then cc.error('SImg_BgApprox2: Invalid image: fi');

    for c:=0 to fi.nc-1 do
        res[c]:=simg_large;

    SImg_Filter(cc, fi,fj, 4.5, -2, 9, 0.5);
    for y:=0 to fj.ny-2 do
    begin
        d1:=fj.pt + y*fj.sl;
        d2:=d1 + fj.nc;
        d3:=d1 + fj.sl;
        d4:=d2 + fj.sl;
        for x:=0 to fj.nx-2 do
            for c:=0 to fj.nc-1 do
            begin
                r:=d1^+d2^+d3^+d4^;
                d1+=1; d2+=1; d3+=1; d4+=1;
                if r<res[c] then res[c]:=r;
            end;
    end;
    SImg_Free(cc,fj);

    for c:=0 to fi.nc-1 do
        res[c]*=0.25;

    SImg_DecDebugLevel;
end;


{ constant background approximation and subtraction }
procedure SImg_BgApprox2(cc:CSImg_Config; fi:TSImg_FPImage);
var res,d : ^simgfp;
    i,c   : longint;
begin
    getmem(res,fi.nc*sizeof(simgfp));
    SImg_BgApprox2(cc,fi,res);
    d:=fi.pt;
    for i:=0 to fi.nx*fi.ny-1 do
    begin
        for c:=0 to fi.nc-1 do
            d[c]-=res[c];
        d+=fi.nc;
    end;
    freemem(res,fi.nc*sizeof(simgfp));
end;

{*************************************************************************}
{******* SImg_Approx *****************************************************}
{*************************************************************************}
{ calculate scalar products
    ya,yz     : region for this thread
    n         : highest order of radial polynomials
    sp        : scalar products (results)
    a         : factors of previous solutions
    dev       : deviation
   polynomials:
    p_0(x,y)=x
    p_1(x,y)=y
    p_2(x,y)=x*y
    p_(3+i)=(x^2+y^2)^i
    p_0 to p_2 are orthogonal to all other polynomials
}
threadprocedure SImg_approx_thread1(fi:TSImg_FPImage; ya:longint; yz:longint; n:longint; sp:pdouble; a:pdouble );
var i,j,k,c     : longint;
    x,y,r,s     : double;
    p3,pr       : double;
    d           : ^simgfp;
    e,f         : ^double;
begin
    for i:=0 to fi.nc*(n+4)-1 do
        sp[i]:=0;

    d:=fi.pt + ya*fi.sl;
    for i:=ya to yz do
    begin
        y:=i-(fi.ny-1)*0.5;
        for j:=0 to fi.nx-1 do
        begin
            x:=j-(fi.nx-1)*0.5;
            p3:=x*y;
            pr:=sqr(x)+sqr(y);
            e:=sp;
            f:=a;
            for c:=0 to fi.nc-1 do
            begin
                { cap limit }
                s:=f[0]*x + f[1]*y +  f[2]*p3 + f[3];   f+=4;
                r:=1;
                for k:=0 to n-1 do
                begin
                    r*=pr;
                    s+=f[k]*r;
                end;
                f+=n;
                r:=d^;  d+=1;
                if r>s then r:=s;
                { scalar product }
                e[0] += r*x;
                e[1] += r*y;
                e[2] += r*p3;
                e[3] += r;
                e+=4;
                for k:=0 to n-1 do
                begin
                    r*=pr;
                    e[k]+=r;
                end;
                e+=n;
            end;
        end;
    end;
end;

{ calculate scalar products
    ya,yz     : region for this thread
    n         : highest order of radial polynomials
    a         : factors of solutions
    dm        : difference mode
    ss        : step size
   polynomials:
    p_0(x,y)=x
    p_1(x,y)=y
    p_2(x,y)=x*y
    p_(3+i)=(x^2+y^2)^i
}
threadprocedure SImg_approx_thread2(fi:TSImg_FPImage; ya:longint; yz:longint; n:longint; a:pdouble; dm:boolean; ss:longint );
var i,j,k,c   : longint;
    x,y,xc,yc : double;
    r,s,rs    : double;
    p3,pr     : double;
    f1,f2     : double;
    d         : ^simgfp;
    f         : ^double;
begin
    d:=fi.pt + ya*fi.sl;
    if dm then
    begin
        f1:=1;
        f2:=-1;
    end
    else begin
        f1:=0;
        f2:=1;
    end;
    if ss<1 then ss:=1;
    rs:=1/ss;
    xc:=((fi.nx-1) div ss)*0.5 + (((fi.nx-1) mod ss) div 2)*rs;
    yc:=((fi.ny-1) div ss)*0.5 + (((fi.ny-1) mod ss) div 2)*rs;
    for i:=ya to yz do
    begin
        y:=i*rs-yc;
        for j:=0 to fi.nx-1 do
        begin
            x:=j*rs-xc;
            p3:=x*y;
            pr:=sqr(x)+sqr(y);
            f:=a;
            for c:=0 to fi.nc-1 do
            begin
                r:=f[0]*x + f[1]*y +  f[2]*p3 + f[3];   f+=4;
                s:=1;
                for k:=0 to n-1 do
                begin
                    s*=pr;
                    r+=f[k]*s;
                end;
                f+=n;
                d^:=d^*f1 + f2*r;  d+=1;
            end;
        end;
    end;
end;

{ calculate minima
    ya,yz     : region for this thread
    n         : highest order of radial polynomials
    buf       : output buffers
    a         : factors of previous solutions
   polynomials:
    p_0(x,y)=x
    p_1(x,y)=y
    p_2(x,y)=x*y
    p_(3+i)=(x^2+y^2)^i
    p_0 to p_2 are orthogonal to all other polynomials
}
threadprocedure SImg_approx_thread3(fi:TSImg_FPImage; ya:longint; yz:longint; n:longint; buf:pdouble; a:pdouble );
var i,j,k,c  : longint;
    x,y,r,s  : double;
    p3,pr    : double;
    d        : ^simgfp;
    f        : ^double;
begin
    for i:=0 to fi.nc-1 do
        buf[i]:=simg_large;

    d:=fi.pt + ya*fi.sl;
    for i:=ya to yz do
    begin
        y:=i-(fi.ny-1)*0.5;
        for j:=0 to fi.nx-1 do
        begin
            x:=j-(fi.nx-1)*0.5;
            p3:=x*y;
            pr:=sqr(x)+sqr(y);
            f:=a;
            for c:=0 to fi.nc-1 do
            begin
                s:=f[0]*x + f[1]*y +  f[2]*p3 + f[3];   f+=4;
                r:=1;
                for k:=0 to n-1 do
                begin
                    r*=pr;
                    s+=f[k]*r;
                end;
                f+=n;
                s:=d^-s;
                if s<buf[c] then buf[c]:=s;
                d+=1;
            end;
        end;
    end;
end;

{ approximate an image by polynomials
    rr   : Factor for radial polynomial
    r1   : Factor for 1st order polynomial (x and y)
    r1b  : Factor for 1st order bi-polynomial (x*y)
    n    : highest order of radial polynomials ((x-xc)^2+(y-yc)^2)^i
    it   : background approximation iterations (cap values to previous approximation), strict mode if >5
    dm   : false: approximation replaces input
           true: difference replaces input
    frad : radius median pre-filter, <=0 to disable it
   polynomials:
    p_0(x,y)=x
    p_1(x,y)=y
    p_2(x,y)=x*y
    p_(3+i)=(x^2+y^2)^i
    p_0 to p_2 are orthogonal to all other polynomials
}
procedure SImg_Approx(cc:CSImg_Config; fj:TSImg_FPImage; rr,r1,r1b:double; n,it:longint; dm:boolean; frad:double);
var d          : array[0..2] of double;         { diagonal elements d_i = <p_i,p_i>, factor }
    ma,vx,vy   : matrix;
    fi         : TSImg_FPImage;
    xo,yo,x,y  : double;
    pc,s,r     : double;
    sp,a       : ^double;                       { buffer for <p_i,f> and approximation coefficients }
    i,j,k,c,tn : longint;
    ya,yz      : longint;
    irad,ss    : longint;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Approx: fj='+SImg_InfoStr2(fj)+' rr='+real2str(rr)+' r1='+real2str(r1)+' r1b='+real2str(r1b)+' n='+int2str(n)+' it='+int2str(it)+' dm='+boolean2str(dm)+' rfad='+real2str(frad,-1));
    if not SImg_Recover(cc,fj) then cc.error('SImg_Approx: Invalid image: fj');

    ma.pt:=nil;
    vx.pt:=nil;
    vy.pt:=nil;

    { pre-filter }
    fi.pt:=nil;
    irad:=round(frad-0.45);
    if irad<0 then irad:=0;
    ss:=round(sqrt(irad*2+0.3));
    if irad>0 then SImg_Filter(cc, fj,fi, irad+0.5, -2, ss, 0.5)
    else fi:=fj;

    { calculate d[0] }
    xo:=(1+(fi.nx and 1))*0.5;
    d[0]:=0;
    for i:=0 to (fi.nx-2) shr 1 do
        d[0]+=sqr(xo+i);
    d[0]*=2*fi.ny;

    { calculate d[1] }
    yo:=(1+(fi.ny and 1))*0.5;
    d[1]:=0;
    for i:=0 to (fi.ny-2) shr 1 do
        d[1]+=sqr(yo+i);
    d[1]*=2*fi.nx;

    { calculate d[2] and matrix for radial polynomials }
    //rpc:=(sqr(fi.nx)+sqr(fi.ny))*0.125;
    pc:=8/(sqr(fi.nx)+sqr(fi.ny));
    d[2]:=0;
    am(vy,zmatrix(1,2*n+1));
    for i:=0 to (fi.ny-2) shr 1 do
    begin
        y:=yo+i;
        for j:=0 to (fi.nx-2) shr 1 do
        begin
            x:=xo+j;
            d[2]+=sqr(y*x);
            r:=pc*(sqr(x)+sqr(y));
            s:=1;
            for k:=0 to 2*n do
            begin
                vy.pt[k]+=s;
                s*=r;
            end
        end;
    end;
    d[2]*=4;
    am(ma,zmatrix(n+1,n+1));
    for i:=0 to n do
        for j:=0 to n do
            ma.pt[i*(n+1)+j]:=4*vy.pt[i+j];
    am(vy,zmatrix(1,n+1));

    { preparations }
    tn:=cc.threads;
    k:=n+4;
    getmem2(sp, tn*fi.nc*k*sizeof(double));
    getmem2(a, fi.nc*k*sizeof(double));
    for j:=0 to fi.nc-1 do
    begin
        for i:=0 to k-1 do
            a[j*k+i]:=0;
        a[j*k+3]:=simg_large;
    end;

    { approximation iterations }
    for i:=0 to it do
    begin

        { calculate <p_i,f> }
        ya:=0;
        for j:=0 to tn-1 do
        begin
            yz:=fi.ny*(j+1) div tn;
            dothread SImg_Approx_thread1(fi, ya,yz-1, n, sp+j*fi.nc*(n+4), a);
            ya:=yz;
        end;
        threadwait;

        { solve for factors }
        for c:=0 to fi.nc-1 do
        begin
            x:=1;
            for k:=0 to n+3 do
            begin
                y:=0;
                for j:=0 to tn-1 do
                    y+=sp[ (j*fi.nc+c)*(n+4) + k];
                case k of
                    0 : a[c*(n+4)+0]:=y*r1/d[0];
                    1 : a[c*(n+4)+1]:=y*r1/d[1];
                    2 : a[c*(n+4)+2]:=y*r1b/d[2];
                    else begin
                        vy.pt[k-3]:=y*x;
                        x*=pc;
                    end;
                end;
            end;
            am(vx,solve(ma,vy));
            if i=it then x:=rr
            else x:=1;
            for k:=0 to n do
            begin
                a[c*(n+4)+3+k]:=vx.pt[k]*x;
                x*=pc;
            end;
        end;

{        for k:=0 to n+3 do
            write(stderr,real2str(a[c*(n+4)+k]),'   ');
        writeln(stderr); }
    end;

    { find minima for strict background mode }
    if it>5 then
    begin
        ya:=0;
        for j:=0 to tn-1 do
        begin
            yz:=fi.ny*(j+1) div tn;
            dothread SImg_Approx_thread3(fi, ya,yz-1, n, sp+j*fi.nc, a);
            ya:=yz;
        end;
        threadwait;
        for k:=0 to fi.nc-1 do
        begin
            r:=sp[k];
            for j:=1 to tn-1 do
                if sp[j*fi.nc+k]<r then r:=sp[j*fi.nc+k];
            a[k*(n+4)+3]+=r;
//            write(stderr,'  ',real2str(r));
        end;
//        writeln(stderr);
    end;

    if irad>0 then SImg_free(cc,fi);
    free(ma);
    free(vx);
    free(vy);

    ya:=0;
    for j:=0 to tn-1 do
    begin
        yz:=fj.ny*(j+1) div tn;
        dothread SImg_Approx_thread2(fj, ya,yz-1, n, a, dm, ss);
        ya:=yz;
    end;
    threadwait;

    freemem2(sp, tn*fi.nc*(n+4)*sizeof(double));
    freemem2(a, fi.nc*(n+4)*sizeof(double));
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_BGApprox3 ******************************************************}
{*******************************************************************************}
//ifdef[bgapprox3_debug]
const SImg_BGApprox3_first : boolean = true;
//endif
{ Background subtraction using polynomials and image
    fi: input and background output (if fg is background subtracted
    fb: background approximation image (optional)
    fp: polynomial imge which is multiplied with the polynomials such that DOF are doubled, optional
    fw: pixel weight, optional
    fg: Gradient output if og=true, otherwise ignored
    bg: Approximation method:
        1: constant
        2: gradient
        3: curved
        4+: 1+ radial polynomials
    dm: subtract background from fi
    og: output gradient to fg
    tm:  True mosaic mode: equalize weight
    pf: factor control, if not nil:
         input:  vector of factors for approximation functions (one vector per color)
         output: approximation factors (one vector per color)
}
procedure SImg_BGApprox3(cc:CSImg_Config; fi,fb,fp,fw:TSImg_FPImage; var fg:TSImg_FPImage; bg:longint; dm,og,tm:boolean; pf:pdouble);
const f_ss         = 4;          { down-sampling. Filter diameter is 2*f_ss+1 }
      f_rad        = f_ss*1.5;   { median filter radius }
      weight_fact  = 0.8;        { weight at noise floor }
      weight_limit = 5;          { weight limit, >1 }
      iterations   = 10;

var fi2,fw2,fw3,fj,fb2,fp2,fw1  : TSImg_FPImage;
    x,y,c,xo,yo,x2,y2,i,j       : longint;
    d,e,f,g                     : ^simgfp;
    a0,a1,a2,a3                 : simgfp;
    r,s,t,u,v                   : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BGApprox3: Called with: fi='+SImg_InfoStr2(fi)+' fb='+SImg_InfoStr2(fb)+' fp='+SImg_InfoStr2(fp)+' fw='+SImg_InfoStr2(fw)+' fg='+SImg_InfoStr2(fw)+' bg='+int2str(bg)+' dm='+boolean2str(dm)+' og='+boolean2str(og)+' tm='+boolean2str(tm));
    if bg>1 then bg+=1; // degrees of freedom

    if not SImg_Recover(cc,fi) then cc.error('SImg_BGApprox3: Invalid image: fi');
    if fw.pt<>nil then
    begin
        if not SImg_Recover(cc,fw) then cc.error('SImg_BGApprox3: Invalid image: fw');
        if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>fi.nc) then cc.error('SImg_BGApprox3: Incompatible dimensions of fw and fi');
    end;
    if fb.pt<>nil then
    begin
        if not SImg_Recover(cc,fb) then cc.error('SImg_BGApprox3: Invalid image: fb');
        if (fb.nx<>fi.nx) or (fb.ny<>fi.ny) or (fb.nc<>fi.nc) then cc.error('SImg_BGApprox3: Incompatible dimensions of fb and fi');
    end;
    if fp.pt<>nil then
    begin
        if not SImg_Recover(cc,fp) then cc.error('SImg_BGApprox3: Invalid image: fp');
        if (fp.nx<>fi.nx) or (fp.ny<>fi.ny) or (fp.nc<>fi.nc) then cc.error('SImg_BGApprox3: Incompatible dimensions of fp and fi');
    end;
    if og and (not SImg_OverWriteAlert(fi,fg)) and (not SImg_OverWriteAlert(fb,fg)) and  (not SImg_OverWriteAlert(fp,fg)) and (not SImg_OverWriteAlert(fw,fg)) then SImg_Free(cc,fg);

    { Down-scaling by factor 2 (bayer compatible) + sub-sample and median filter
      No weighted averaging because that causes artifacts due to photon noise. }
    fi2.pt:=nil; fw1.pt:=nil; fb2.pt:=nil; fp2.pt:=nil;
    SImg_New(cc, fi2, fi.nx div 2, fi.ny div 2, fi.nc, false);
    if fb.pt<>nil then SImg_New(cc, fb2, fi2.nx, fi2.ny, fi2.nc, false);
    if fp.pt<>nil then SImg_New(cc, fp2, fi2.nx, fi2.ny, fi2.nc, false);
    SImg_New(cc, fw1, fi2.nx, fi2.ny, fi2.nc, false);
    xo:=((fi.nx and 1)+1) div 2; { round to opposite direction as SImg_Filter }
    yo:=((fi.ny and 1)+1) div 2;
    j:=0;
    for y:=0 to fi2.ny-1 do
    begin
        y2:=(y*2+yo)*fi.sl + xo*fi.nc;
        for x:=0 to fi2.nx-1 do
        begin
            x2:=y2 + x*2*fi.nc;
            d:=fi.pt + x2;
            e:=fw.pt + x2;
            f:=fb.pt + x2;
            g:=fp.pt + x2;
            for c:=0 to fi.nc-1 do
            begin
                if fw.pt<>nil then
                begin
                    a0:=e[0];     a1:=e[      fw.nc]; if a1<a0 then a0:=a1;
                    a2:=e[fw.sl]; a3:=e[fw.sl+fw.nc]; if a3<a2 then a2:=a3;
                    if a2<a0 then fw1.pt[j]:=a2
                    else fw1.pt[j]:=a0;
                    e+=1;
                end
                else fw1.pt[j]:=1;
                fi2.pt[j]:=0.25*(d[0] + d[fi.nc] + d[fi.sl] + d[fi.sl+fi.nc]);
                d+=1;
                if fb2.pt<>nil then
                begin
                    fb2.pt[j]:=0.25*(f[0] + f[fb.nc] + f[fb.sl] + f[fb.sl+fb.nc]);
                    f+=1;
                end;
                if fp2.pt<>nil then
                begin
                    fp2.pt[j]:=0.25*(g[0] + g[fp.nc] + g[fp.sl] + g[fp.sl+fp.nc]);
                    g+=1;
                end;
                j+=1;
            end;
        end;
    end;
    fw3.pt:=nil;
    SImg_New(cc,fw3, ((fi2.nx-1) div f_ss)+1, ((fi2.ny-1) div f_ss)+1, fi2.nc, false);  { sub-sampled backup of fi2 for noise estimation }
    xo:=((fi2.nx-1) mod f_ss) div 2;
    yo:=((fi2.ny-1) mod f_ss) div 2;
    for y:=0 to fw3.ny-1 do
    begin
        d:=fi2.pt + (y*f_ss+yo)*fi2.sl + xo*fi2.nc;
        e:=fw3.pt + y*fw3.sl;
        for x:=0 to fw3.nx-1 do
        begin
            for c:=0 to fi2.nc-1 do
                e[c]:=d[c];
            d+=f_ss*fi.nc;
            e+=fi2.nc;
        end;
    end;
    SImg_Filter(cc, fi2,fi2, f_rad, -4, f_ss, 0.5);
    if fb2.pt<>nil then SImg_Filter(cc, fb2,fb2, f_rad, -4, f_ss, 0.5);
    if fp2.pt<>nil then SImg_Filter(cc, fp2,fp2, f_rad, -4, f_ss, 0.5);
    SImg_Filter(cc, fw1,fw1, f_rad, -4, f_ss, 0.35);
    fw2.pt:=nil;
    if tm then SImg_EqualizeWeight_int(cc,fw1,fw2)
    else fw2:=fw1;
//ifdef[bgapprox3_debug]
    if SImg_BGApprox3_first then
    begin
//      SImg_WriteImageFile(cc,'bgApprox3-fi2v.fits',fw3);
        SImg_WriteImageFile(cc,'bgApprox3-fi2.fits',fi2);
        SImg_WriteImageFile(cc,'bgApprox3-fw1.fits',fw1);
        if fw2.pt<>fw1.pt then SImg_WriteImageFile(cc,'bgApprox3-fw2.fits',fw2);
        if fb2.pt<>nil then SImg_WriteImageFile(cc,'bgApprox3-fb2.fits',fb2);
        if fp2.pt<>nil then SImg_WriteImageFile(cc,'bgApprox3-fp2.fits',fp2);
    end;
//endif

    { estimate noise from difference of input and output of median filter, initialize fw3 }
    t:=simg_mina; s:=0;
    for i:=0 to fw3.size-1 do
    begin
        r:=fw2.pt[i];
        if r<0 then r:=0;
        t+=abs(fi2.pt[i]-fw3.pt[i])*sqrt(fw1.pt[i])*r;
        s+=r;
        fw3.pt[i]:=r;
    end;
//    cc.info('sig='+real2str(t/s));
    u:=sqr(s*1.65*f_rad/t) * ln(weight_fact); { sqr(t/(s*1.65*f_rad)) is estimated E[sqr(fi2)*fw2],  1.65 = sqrt(pi) / (0.5/0.47) , determined using bgapprox_test }
    v:=1/(weight_limit-1);

    { calculate weight }
    fj.pt:=nil;
    SImg_Duplicate(cc, fi2,fj, true);
    for i:=1 to iterations do
    begin
        SImg_ApproxW_int(cc, fj,fw3,fp2,fb2, bg, true, nil);
        for x:=0 to fw2.size-1 do
        begin
            r:=fj.pt[x];
            s:=fw2.pt[x];
            t:=sqr(r)*fw1.pt[x]*u;                              { u<0 }
            if r>0 then fw3.pt[x]:=s*exp(t)
            else fw3.pt[x]:=s*(1-t/(1-t*v));
        end;
//      SImg_WriteImageFile(cc,'bgApprox3-fw3-'+int2str(i)+'.fits',fw3);
    end;
    SImg_Free(cc,fj);
    SImg_Free(cc,fw1);
    SImg_Free(cc,fw2);

    { calculate gradient }
    SImg_ApproxW_int(cc, fi2,fw3,fp2,fb2, bg, false, pf);
//ifdef[bgapprox3_debug]
    if SImg_BGApprox3_first then SImg_WriteImageFile(cc,'bgApprox3-fw3.fits',fw3);
    if SImg_BGApprox3_first then SImg_WriteImageFile(cc,'bgApprox3-fg.fits',fi2);
//endif
    SImg_Free(cc,fb2);
    SImg_Free(cc,fp2);
    SImg_Free(cc,fw3);

    { output }
    if og then SImg_New(cc, fj, fi.nx,fi.ny,fi.nc, false);
    if og or dm then
    begin
        i:=0;
        xo:=(fi2.nx*f_ss*2 - fi.nx) div 2;
        yo:=(fi2.ny*f_ss*2 - fi.ny) div 2;
        for y:=0 to fi.ny-1 do
        begin
            y2:=(y-yo) div (f_ss*2);
            for x:=0 to fi.nx-1 do
            begin
                e:=fi2.pt + y2*fi2.sl + ((x-xo) div (f_ss*2))*fi2.nc; { no range check required }
                if dm then
                    for c:=0 to fi.nc-1 do
                        fi.pt[i+c]-=e[c];
                if og then
                    for c:=0 to fi.nc-1 do
                        fj.pt[i+c]:=e[c];
                i+=fi.nc;
            end;
        end;
    end;
    if og then SImg_Move(cc, fj,fg);
//ifdef[bgapprox3_debug]
//    if SImg_BGApprox3_first then SImg_WriteImageFile(cc,'bgApprox3-fi.fits',fi);
    SImg_BGApprox3_first:=false;
//endif
    SImg_Free(cc,fi2);

    SImg_DecDebugLevel;
end;

procedure SImg_BGApprox3(cc:CSImg_Config; fi,fb,fp,fw:TSImg_FPImage; bg:longint; tm:boolean);   { dm=true, og=false, pf=nil }
begin
    SImg_BGApprox3(cc, fi,fb,fp,fw, fi{will not be touched}, bg, true,false,tm, nil);
end;

function SImg_BGApprox3(cc:CSImg_Config; fi,fb,fw:TSImg_FPImage; var fg:TSImg_FPImage; og:boolean):double;      { bg=2, dm=false, tm=false, pf[0..2]=1, pf[3]=0, result: factor for fb }
var fe  : TSImg_FPImage;
    pf  : ^double;
    i   : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_BGApprox3: Invalid image: fi');
    getmem(pf, fi.nc*sizeof(double)*4);
    for i:=0 to fi.nc*4-1 do
        if i and 3=3 then pf[i]:=0
        else pf[i]:=1;
    fe.pt:=nil;
    SImg_BGApprox3(cc, fi,fb,fe,fw, fg, 2, false,og,false, pf);
    result:=0;
    for i:=0 to fi.nc-1 do
        result+=pf[i*4+3];
    result/=fi.nc;
    freemem(pf, fi.nc*sizeof(double)*4);
end;



{*******************************************************************************}
{********* SImg_Imgapprox ******************************************************}
{*******************************************************************************}
//ifdef[approx_debug]
const SImg_Imgapprox_first : boolean = true;
//endif
{ Approximate fi by fa and polynomials and return the difference.
    fi:  Input image and background corrected output (1 color)
    fa:  Reference image image (1 colors)
    fb:  background approximation image (optional)
    fp:  polynomial image which is multiplied with the polynomials such that DOF are doubled, optional
    fo1: Residual output
    fo2: Background corrected output
    bg:  polynomial approximation method:
         0: none
         1: constant
         2: gradient
         3: curved
         4+: 1+ radial polynomials
    tm:  True mosaic mode: equalize weight
    result: Approximation factor of reference image
    calibration factor (1 if bc=false)
}
function SImg_ImgApprox(cc:CSImg_Config; fi,fa,fb,fp,fw:TSImg_FPImage; var fo1,fo2:TSImg_FPImage; bg:longint; tm:boolean):double;
const f_ss      = 4;            { down-sampling }
      f_rad     = f_ss*1.5;     { filter radius }
      q_um_low  = 1e-3;         { lower unmask quantile for gradient calculation }
      q_um_high = 1-5e-3;       { higher unmask quantile for gradient calculation }

var fi2,fa2,fw2,fi1,fa1,fb2,fp2 : TSImg_FPImage;
    x,y,i,xo,yo,x2,y2           : longint;
    a,b,r,s,t                   : double;
    d,e                         : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Imgapprox: Called with: fi='+SImg_InfoStr2(fi)+' fa='+SImg_InfoStr2(fa)+' fb='+SImg_InfoStr2(fb)+' fp='+SImg_InfoStr2(fp)+' fw='+SImg_InfoStr2(fw)+' fo1='+SImg_InfoStr2(fo1)+' fo2='+SImg_InfoStr2(fo2)+' bg='+int2str(bg)+' tm='+boolean2str(tm));
    if bg>1 then bg+=1; // degrees of freedom

    if not SImg_Recover(cc,fi) then cc.error('SImg_Imgapprox: Invalid image: fi');
    if not SImg_Recover(cc,fa) then cc.error('SImg_Imgapprox: Invalid image: fa');
    if not SImg_Recover(cc,fw) then cc.error('SImg_Imgapprox: Invalid image: fw');
    if (fi.nc<>1) then cc.error('SImg_Imgapprox: Invalid dimensions fi');
    if (fa.nx<>fi.nx) or (fa.ny<>fi.ny) or (fa.nc<>1) then cc.error('SImg_Imgapprox: Invalid dimensions of fa');
    if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>1) then cc.error('SImg_Imgapprox: Invalid dimensions of fw');
    if fb.pt<>nil then
    begin
        if not SImg_Recover(cc,fb) then cc.error('SImg_ImgApprox: Invalid image: fb');
        if (fb.nx<>fi.nx) or (fb.ny<>fi.ny) or (fb.nc<>1) then cc.error('SImg_ImgpAprox: Incompatible dimensions of fb and fi');
    end;
    if fp.pt<>nil then
    begin
        if not SImg_Recover(cc,fp) then cc.error('SImg_ImgApprox: Invalid image: fp');
        if (fp.nx<>fi.nx) or (fp.ny<>fi.ny) or (fp.nc<>1) then cc.error('SImg_ImgpAprox: Incompatible dimensions of fp and fi');
    end;

    if not SImg_OverWriteAlert(fo1,fi) and not SImg_OverWriteAlert(fo1,fa) and not SImg_OverWriteAlert(fo1,fb) and not SImg_OverWriteAlert(fo1,fp) and not SImg_OverWriteAlert(fo1,fw) then
    begin
        SImg_Free(cc,fo1);
        fo1.pt:=nil;
    end;
    if not SImg_OverWriteAlert(fo2,fi) and not SImg_OverWriteAlert(fo2,fa) and not SImg_OverWriteAlert(fo2,fb) and not SImg_OverWriteAlert(fo2,fp) and not SImg_OverWriteAlert(fo2,fw) then
    begin
        SImg_Free(cc,fo2);
        fo2.pt:=nil;
    end;

    { linear down-scaling + filtering
      These operations must be linear and un-weighted in order to avoid effects caused by different PSF width's and by overweighting dark areas }
    fa2.pt:=nil; SImg_Filter(cc, fa, fa2,  f_rad, -3, f_ss);
    fi2.pt:=nil; SImg_Filter(cc, fi, fi2,  f_rad, -3, f_ss);
    fw2.pt:=nil; SImg_Filter(cc, fw, fw2,  f_rad, -5, f_ss);
    if tm then SImg_EqualizeWeight_int(cc,fw2,fw2);
//ifdef[approx_debug]
    if SImg_Imgapprox_first then
    begin
        SImg_WriteImageFile(cc,'Imgapprox-fw.fits',fw);
        SImg_WriteImageFile(cc,'Imgapprox-fi2.fits',fi2);
        SImg_WriteImageFile(cc,'Imgapprox-fw2.fits',fw2);
        SImg_WriteImageFile(cc,'Imgapprox-fa2.fits',fa2);
    end;
//endif
    fb2.pt:=nil;
    if fb.pt<>nil then
    begin
        SImg_Filter(cc, fb, fb2,  f_rad, -3, f_ss);
//ifdef[approx_debug]
        if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fb2.fits',fb2);
//endif
    end;
    fp2.pt:=nil;
    if fp.pt<>nil then
    begin
        SImg_Filter(cc, fp, fp2,  f_rad, -3, f_ss);
//ifdef[approx_debug]
        if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fp2.fits',fp2);
//endif
    end;

    { polynomial and image approximation }
    fi1.pt:=nil; fa1.pt:=nil;
    SImg_Duplicate(cc, fi2,fi1, true);
    SImg_Duplicate(cc, fa2,fa1, true);
    SImg_ApproxW_int(cc, fi2,fw2,fp2,fb2, bg, true, nil);
    SImg_ApproxW_int(cc, fa2,fw2,fp2,fb2, bg, true, nil);

    { reference image approximation }
    a:=simg_mina; b:=simg_mina;
    for x:=0 to fi2.size-1 do
    begin
        t:=fw2.pt[x];
        if t<0 then t:=0;
        r:=fa2.pt[x];
        s:=fi2.pt[x];
        a+=sqr(r)*t;
        b+=r*s*t;
    end;
    b/=a;
    cc.info(cc.name+': Gradient and background subtraction. Image approximation factor: '+real2str(b,-3));
    if b>simg_mina then result:=1/b
    else result:=1;

    { unmask large residuals }
    for x:=0 to fi2.size-1 do
        fi2.pt[x]-=b*fa2.pt[x];         { residual }
//ifdef[approx_debug]
    if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fr2.fits',fi2);
//endif
    s:=SImg_GetQuantile(cc, fi2, q_um_low);
    t:=SImg_GetQuantile(cc, fi2, q_um_high);
    for x:=0 to fw2.size-1 do
    begin
        r:=fi2.pt[x];
        if (r<s) or (r>t) then fw2.pt[x]:=0;
    end;
//ifdef[approx_debug]
    if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fw3.fits',fw2);
//endif

    { gradients for fo1 and fo2 }
    a:=1/result;
    for x:=0 to fi2.size-1 do
    begin
        r:=fi1.pt[x];
        s:=fa1.pt[x];
        t:=r - b*s;                     { for gradient calculation for fo1 }
        fa1.pt[x]:=t;  fi1.pt[x]:=t;
        t:=r - a*s;                     { for gradient calculation for fo2 }
        fa2.pt[x]:=t;  fi2.pt[x]:=t;
    end;
    SImg_ApproxW_int(cc, fi1,fw2,fp2,fb2, bg, true, nil);
    SImg_ApproxW_int(cc, fi2,fw2,fp2,fb2, bg, true, nil);
    for i:=0 to fa1.size-1 do
    begin
        fa1.pt[i]-=fi1.pt[i];
        fa2.pt[i]-=fi2.pt[i];
    end;

    { output }
    SImg_Free(cc, fi1);
    SImg_Free(cc, fi2);
    SImg_Free(cc, fw2);
    SImg_Free(cc, fb2);
    SImg_Free(cc, fp2);

    if fo1.pt=nil then SImg_New(cc, fo1, fi.nx,fi.ny,1, false);
    if fo2.pt=nil then SImg_New(cc, fo2, fi.nx,fi.ny,1, false);
    d:=fo1.pt;
    e:=fo2.pt;
    i:=0;

    xo:=(fi2.nx*f_ss - fi.nx) div 2;
    yo:=(fi2.ny*f_ss - fi.ny) div 2;
    for y:=0 to fi.ny-1 do
    begin
        y2:=(y-yo) div f_ss;
        for x:=0 to fi.nx-1 do
        begin
            x2:=(x-xo) div f_ss + fi2.sl*y2;    { no range check required }
            r:=fi.pt[i];
            s:=fa.pt[i];
            d[i]:=r - b*s - fa1.pt[x2];
            e[i]:=r       - fa2.pt[x2];
            i+=1;
        end;
    end;

//ifdef[approx_debug]
    if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fg1.fits',fa1);
    if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fg2.fits',fa2);
//    if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fo.fits',fo2);
    if SImg_Imgapprox_first then SImg_WriteImageFile(cc,'Imgapprox-fres.fits',fo1);
    SImg_Imgapprox_first:=false;
//endif
    SImg_Free(cc, fa1);
    SImg_Free(cc, fa2);

    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_ApproxMosaic ***************************************************}
{*******************************************************************************}
{ Approximate fi by polynomials. Border will be overweighted (mosaic mode)
    fi: Input and output background corrected output
    fb: background approximation image (optional)
    fp: polynomial imgage which is multiplied with the polynomials such that DOF are doubled, optional
    fw: Weight
    bg: polynomial approximation method:
         0: none
         1: constant
         2: gradient
         3: curved
         4+: 1+ radial polynomials
    mosaic: boundary overweight factor, >0
    tm:  True mosaic mode: equalize weight
}
procedure SImg_ApproxMosaic(cc:CSImg_Config; fi,fb,fp,fw:TSImg_FPImage; bg:longint; mosaic:double; tm:boolean);
const q_low     = 1e-3;         // lower unmask quantile for gradient calculation
      q_high    = 1-5e-3;       // higher unmask quantile for gradient calculation
      w_fact    = 1.0;          // weight is limited to weighted weight multiplied by this parameter
var x,y,i       : longint;
    r,s,t,u     : double;
    fw2         : TSImg_FPImage;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_ApproxMosaic: Called with: fi='+SImg_InfoStr2(fi)+' fb='+SImg_InfoStr2(fb)+' fp='+SImg_InfoStr2(fp)+' fw='+SImg_InfoStr2(fw)+' bg='+int2str(bg)+' mosaic='+real2str(mosaic)+' tm='+boolean2str(tm));
    if not SImg_Recover(cc,fi) then cc.error('SImg_ApproxMosaic: Invalid image: fi');
    if not SImg_Recover(cc,fw) then cc.error('SImg_ApproxMosaic: Invalid image: fw');
    if (fi.nc<>1) then cc.error('SImg_ApproxMosaic: Invalid dimensions fi');
    if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>1) then cc.error('SImg_ApproxMosaic: Invalid dimensions of fw');
    if fb.pt<>nil then
    begin
        if not SImg_Recover(cc,fb) then cc.error('SImg_ImgApprox: Invalid image: fb');
        if (fb.nx<>fi.nx) or (fb.ny<>fi.ny) or (fb.nc<>1) then cc.error('SImg_ImgpAprox: Incompatible dimensions of fb and fi');
    end;
    if fp.pt<>nil then
    begin
        if not SImg_Recover(cc,fp) then cc.error('SImg_ImgApprox: Invalid image: fp');
        if (fp.nx<>fi.nx) or (fp.ny<>fi.ny) or (fp.nc<>1) then cc.error('SImg_ImgpAprox: Incompatible dimensions of fp and fi');
    end;
    if bg>1 then bg+=1; // degrees of freedom
    if mosaic<0 then mosaic:=0;

    { unmask extreme values }
    s:=SImg_GetQuantile(cc, fi, q_low);
    t:=SImg_GetQuantile(cc, fi, q_high);
    fw2.pt:=nil;
    SImg_New(cc, fw2, fw.nx,fw.ny,1, false);
    for x:=0 to fw.size-1 do
    begin
        r:=fi.pt[x];
        if (r<s) or (r>t) then fw2.pt[x]:=0
        else fw2.pt[x]:=fw.pt[x];
    end;

    if tm then SImg_EqualizeWeight_int(cc,fw2,fw2);

    { boundary overweight }
    if mosaic>1e-3 then
    begin
        mosaic/=1+mosaic;
        s:=2/(fw2.nx-1);
        t:=2/(fw2.ny-1);
        i:=0;
        for y:=0 to fw2.ny-1 do
        begin
            r:=y*t;
            r*=(2-r)*mosaic;
            for x:=0 to fw2.nx-1 do
            begin
                u:=x*s;
                fw2.pt[i]*=1-u*(2-u)*r;
                i+=1;
            end;
        end;
    end;

    { approximation }
//    SImg_WriteImageFile(cc,'approxMosaic-fw2.fits',fw2);
//    SImg_WriteImageFile(cc,'approxMosaic-fi.fits',fi);
    SImg_ApproxW_int(cc, fi,fw2,fp,fb, bg, true, nil);
//    SImg_WriteImageFile(cc,'approxMosaic-fj.fits',fi);
//    cc.error('no error');

    SImg_Free(cc,fw2);
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_Imgapprox2 *****************************************************}
{*******************************************************************************}
{ Approximate fi by fa. Return the factor for fa.
    fi:  Input image and background corrected output (1 color)
    fa:  Approximation image (1 colors)
    fw:  Weight for approximation image
    result: factor for fa
}
function SImg_ImgApprox2(cc:CSImg_Config; fi,fa,fw:TSImg_FPImage):double;
var r,s,t,b0,b1 : double;
    a00,a01,a11 : double;
    i           : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Imgapprox2: Called with: fi='+SImg_InfoStr2(fi)+' fa='+SImg_InfoStr2(fa)+' fw='+SImg_InfoStr2(fw));

    if not SImg_Recover(cc,fi) then cc.error('SImg_Imgapprox2: Invalid image: fi');
    if not SImg_Recover(cc,fa) then cc.error('SImg_Imgapprox2: Invalid image: fa');
    if not SImg_Recover(cc,fw) then cc.error('SImg_Imgapprox2: Invalid image: fw');
    if fa.size<>fi.size then cc.error('SImg_Imgapprox2: Invalid dimensions of fa');
    if fw.size<>fi.size then cc.error('SImg_Imgapprox2: Invalid dimensions of fw');

    a00:=0; a01:=0; a11:=0;
    b0:=0; b1:=0;
    for i:=0 to fi.size-1 do
    begin
        r:=fi.pt[i];
        s:=fw.pt[i];
        t:=fa.pt[i];
        b0+=t*r*s;
        b1+=r*s;
        a00+=sqr(t)*s;
        a01+=t*s;
        a11+=s;
    end;
    r:=a00*a11-sqr(a01);
    if abs(r)<simg_mina then result:=0
    else result:=(a11*b0-a01*b1)/r;

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_BgApprox5 **************************************************}
{*************************************************************************}
{ constant background approximation. Pixels with a weight less than thr*<weighted average weight> are set to weighted average.
    c      : color
    rad    : median filter radius
    result : result buffer for at least fi.nc values
}
function SImg_BgApprox5(cc:CSImg_Config; fi,wi:TSImg_FPImage; c:longint; rad:single):double;
const thr = 0.25;       { ignore threshold, see description above }
var  fj                 : TSImg_FPImage;
     d,e                : ^simgfp;
     i,wc               : longint;
     r,s,sw,sfw,sww     : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BgApprox5; fi='+SImg_InfoStr2(fi)+' wi='+SImg_InfoStr2(fi)+' c='+int2str(c)+' rad='+real2str(rad,-1));
    if not SImg_Recover(cc,fi) then cc.error('SImg_BgApprox5: Invalid image: fi');
    if not SImg_Recover(cc,wi) then cc.error('SImg_BgApprox5: Invalid image: wi');
    if (wi.nx<>fi.nx) or (wi.ny<>fi.ny) or ((wi.nc<>1) and (wi.nc<>fi.nc)) then cc.error('SImg_BGApprox5: Incompatible image dimensions');
    if (c<0) or (c>=fi.nc) then cc.error('SImg_BGApprox5: Invalid color: '+int2str(c));

    if wi.nc=1 then wc:=0
    else wc:=c;
    if rad<1.5 then rad:=1.5;

    { averages, copy }
    fj.pt:=nil;
    SImg_New(cc, fj, fi.nx,fi.ny,1, false);
    sww:=0; sw:=0; sfw:=0;
    d:=fi.pt + c;
    e:=wi.pt + wc;
    for i:=0 to fi.size-1 do
    begin
        r:=e^;  e+=wi.nc;
        s:=d^;  d+=fi.nc;
        sw+=r;
        sww+=sqr(r);
        sfw+=s*r;
        fj.pt[i]:=s;
    end;
    if sw<simg_mina then cc.error('SImg_BGApprox5: Non-positive weight image');
    sw:=1/sw;
    sfw*=sw;
    sww*=sw*thr;

    { mask out pixels }
    e:=wi.pt + wc;
    for i:=0 to fi.size-1 do
    begin
        if e^<sww then fj.pt[i]:=sfw;
        e+=wi.nc;
    end;

    { median filter }
    i:=round(rad*2) div 3;
    SImg_Filter(cc, fj,fj, rad,-2,i, 0.5);

    r:=fj.pt[0];
    for i:=1 to fj.size-1 do
        if fj.pt[i]<r then r:=fj.pt[i];
    SImg_Free(cc, fj);

    result:=r;
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_ImgBGapprox ****************************************************}
{*******************************************************************************}
//ifdef[approx_debug]
const SImg_ImgBGApprox_first : boolean = true;
//endif
{ Approximate background of fi by fb and polynomials and using a reference image
    fi:  Input image and background corrected output (1 color) and background corrected (polynomials only) output
    fa:  Reference image (1 colors)
    fb:  background approximation image (optional)
    fp:  polynomial image which is multiplied with the polynomials such that DOF are doubled, optional
    fw:  Weight for approximation image
    bg:  polynomial approximation method:
         0: none
         1: constant
         2: gradient
         3: curved
         4+: 1+ radial polynomials
    tm:  True mosaic mode: equalize weight
    aa:  Adjust factor for fa and return it. (if false, result is 1)
}
function SImg_ImgBGApprox(cc:CSImg_Config; fi,fa,fb,fp,fw:TSImg_FPImage; bg:longint; tm,aa:boolean):double;
const f_ss      = 4;            { down-sampling }
      f_rad     = f_ss*1.5;     { filter radius }
      bg_it     = 5;            { background weight iterations }
var fi2,fa2,fw2,fb2,fp2 : TSImg_FPImage;
    fa1,fw1,fi1,fe      : TSImg_FPImage;
    x,y,i,xo,yo,y2      : longint;
    a,b,r,s,t           : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_ImgBGApprox: Called with: fi='+SImg_InfoStr2(fi)+' fa='+SImg_InfoStr2(fa)+' fb='+SImg_InfoStr2(fb)+' fp='+SImg_InfoStr2(fp)+' fw='+SImg_InfoStr2(fw)+' bg='+int2str(bg)+' tm='+boolean2str(tm));
    if bg>1 then bg+=1; // degrees of freedom

    if not SImg_Recover(cc,fi) then cc.error('SImg_ImgBGapprox: Invalid image: fi');
    if not SImg_Recover(cc,fa) then cc.error('SImg_ImgBGapprox: Invalid image: fa');
    if not SImg_Recover(cc,fw) then cc.error('SImg_ImgBGapprox: Invalid image: fw');
    if (fi.nc<>1) then cc.error('SImg_ImgBGApprox: Invalid dimensions fi');
    if (fa.nx<>fi.nx) or (fa.ny<>fi.ny) or (fa.nc<>1) then cc.error('SImg_ImgBGApprox: Invalid dimensions of fa');
    if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>1) then cc.error('SImg_ImgBGApprox: Invalid dimensions of fw');
    if fb.pt<>nil then
    begin
        if not SImg_Recover(cc,fb) then cc.error('SImg_ImgBGApprox: Invalid image: fb');
        if (fb.nx<>fi.nx) or (fb.ny<>fi.ny) or (fb.nc<>1) then cc.error('SImg_ImgpAprox: Incompatible dimensions of fb and fi');
    end;
    if fp.pt<>nil then
    begin
        if not SImg_Recover(cc,fp) then cc.error('SImg_ImgBGApprox: Invalid image: fp');
        if (fp.nx<>fi.nx) or (fp.ny<>fi.ny) or (fp.nc<>1) then cc.error('SImg_ImgpAprox: Incompatible dimensions of fp and fi');
    end;

    { linear down-scaling + filtering
      These operations must be linear and un-weighted in order to avoid effects caused by different PSF width's and by overweighting dark areas }
    fa2.pt:=nil; SImg_Filter(cc, fa,fa2,  f_rad, -3, f_ss);
    fi2.pt:=nil; SImg_Filter(cc, fi,fi2,  f_rad, -3, f_ss);
    fw2.pt:=nil; SImg_Filter(cc, fw,fw2,  f_rad, -5, f_ss);
    if tm then SImg_EqualizeWeight_int(cc,fw2,fw2);
//ifdef[approx_debug]
    if SImg_ImgBGApprox_first then
    begin
        SImg_WriteImageFile(cc,'imgBGApprox-fi2.fits',fi2);
        SImg_WriteImageFile(cc,'imgBGApprox-fw2.fits',fw2);
        SImg_WriteImageFile(cc,'imgBGApprox-fa2.fits',fa2);
    end;
//endif
    fb2.pt:=nil;
    if fb.pt<>nil then
    begin
        SImg_Filter(cc, fb,fb2,  f_rad, -3, f_ss);
//ifdef[approx_debug]
        if SImg_ImgBGapprox_first then SImg_WriteImageFile(cc,'imgBGapprox-fb2.fits',fb2);
//endif
    end;
    fp2.pt:=nil;
    if fp.pt<>nil then
    begin
        SImg_Filter(cc, fp,fp2,  f_rad, -3, f_ss);
//ifdef[approx_debug]
        if SImg_ImgBGapprox_first then SImg_WriteImageFile(cc,'imgBGapprox-fp2.fits',fp2);
//endif
    end;

    { background weight }
    fa1.pt:=nil; fw1.pt:=nil;
    SImg_Duplicate(cc, fa2,fa1, true);
    SImg_Duplicate(cc, fw2,fw1, true);

    fe.pt:=nil;
    for i:=1 to bg_it do
    begin
        if bg>0 then SImg_ApproxW_int(cc, fa2,fw2,fe,fe, bg, true, nil);
        a:=simg_mina; b:=0;
        for x:=0 to fa2.size-1 do
        begin
            r:=fw1.pt[x];
            a+=abs(fa2.pt[x])*r;
            b+=r;
        end;
        a:=b/a;
        for x:=0 to fa2.size-1 do
        begin
            r:=1+fa2.pt[x]*a;
            if r>0 then fw2.pt[x]:=fw1.pt[x]/(1+sqr(r))
            else fw2.pt[x]:=fw1.pt[x];
        end;
//ifdef[approx_debug]
        if SImg_ImgBGApprox_first then SImg_WriteImageFile(cc,'imgBGApprox-fw3-'+int2str(i)+'.fits',fw2);
//endif
    end;
    SImg_free(cc, fw1);

    { gradient }
    result:=1;
    if aa then
    begin
        fi1.pt:=nil; SImg_Duplicate(cc, fi2,fi1, true);
        { approximation }
        SImg_ApproxW_int(cc, fa2,fw2,fp2,fb2, bg, true, nil);
        SImg_ApproxW_int(cc, fi2,fw2,fp2,fb2, bg, true, nil);
        { factor for fa }
        a:=simg_mina; b:=simg_mina;
        for x:=0 to fi2.size-1 do
        begin
            t:=fw2.pt[x];
            r:=fa2.pt[x];
            s:=fi2.pt[x];
            a+=sqr(r)*t;
            b+=r*s*t;
        end;
        b/=a;
        result:=b;
        cc.info('Image factor: '+real2str(result));
        { gradient }
        for x:=0 to fi2.size-1 do
            fi2.pt[x]:=fi1.pt[x]-fi2.pt[x] - b*(fa1.pt[x]-fa2.pt[x]);
        SImg_Free(cc,fi1);
    end
    else begin
        for x:=0 to fi2.size-1 do
            fi2.pt[x]-=fa1.pt[x];
        SImg_ApproxW_int(cc, fi2,fw2,fp2,fb2, bg, false, nil);
    end;
    SImg_Free(cc, fa2);
    SImg_Free(cc, fa1);
    SImg_Free(cc, fw2);
    SImg_Free(cc, fb2);
    SImg_Free(cc, fp2);

//ifdef[approx_debug]
    if SImg_ImgBGApprox_first then SImg_WriteImageFile(cc,'imgBGApprox-fg.fits',fi2);
//endif
    i:=0;
    xo:=(fi2.nx*f_ss - fi.nx) div 2;
    yo:=(fi2.ny*f_ss - fi.ny) div 2;
    for y:=0 to fi.ny-1 do
    begin
        y2:=(y-yo) div f_ss;
        for x:=0 to fi.nx-1 do
        begin
            fi.pt[i]-=fi2.pt[(x-xo) div f_ss + fi2.sl*y2];    { no range check required }
            i+=1;
        end;
    end;
//ifdef[approx_debug]
    if SImg_ImgBGApprox_first then SImg_WriteImageFile(cc,'imgBGApprox-fo.fits',fi);
    SImg_ImgBGApprox_first:=false;
//endif
    SImg_Free(cc, fi2);

    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_BGCorrect ******************************************************}
{*******************************************************************************}
{ Background correction using reference image
    fi  : Input image
    fr  : Reference image (optional)
    fo  : output image
    fwhm: PSF FWHM of fi and fr
    ir  : Interpolation ratio
    frad: Filter radius
    fact: factor of input image (relative to reference image). 0: auto-detect
}
procedure SImg_BGCorrect(cc:CSImg_Config; fi,fr:TSImg_FPImage; var fo:TSImg_FPImage; fwhm,ir,frad,fact:double);
const wreg         = 1e-3;      { weight regularization }
      pf : array[0..4] of double = (1,1,1,1,0);
var ps,qr,qi    : array[0..1] of double;
    fw,fg,fj,fk,fl      : TSImg_FPImage;
    rm          : TSImg_RDImage;
    i           : longint;
    r,s         : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BGCorrect: Called with: fi='+SImg_InfoStr2(fi)+' fr='+SImg_InfoStr2(fr)+' fo='+SImg_InfoStr2(fo)+' fwhm='+real2str(fwhm)+' ir='+real2str(ir)+' frad='+real2str(frad));
    if not SImg_Recover(cc,fi) then cc.error('SImg_BGCorrect: Invalid image: fi');
    if (fi.nc<>1) then cc.error('SImg_BGCorrect: Invalid dimensions fi');
    if fr.pt<>nil then
    begin
        if not SImg_Recover(cc,fr) then cc.error('SImg_BGCorrect: Invalid image: fr');
        if (fr.nx<>fi.nx) or (fr.ny<>fi.ny) or (fr.nc<>1) then cc.error('SImg_BGCorrect: Invalid dimensions of fr');
    end;
    if not SImg_OverWriteAlert(fo,fi) and not SImg_OverWriteAlert(fo,fr) then SImg_Free(cc,fo);
    if ir<0.05 then ir:=0.05;
    if ir>0.5 then ir:=0.5;
    if fr.pt=nil then fact:=1;

    { calculate approximation weight }
    ps[1]:=1-ir*0.1;
    ps[0]:=0.25*ps[1];
    SImg_GetQuantile(cc, fi, cc.threads, -1,ps,qi);
    qi[1]-=qi[0];
    if qi[1]<simg_mina then qi[1]:=simg_mina;
    qi[1]:=1/qi[1];
    if fr.pt<>nil then
    begin
        SImg_GetQuantile(cc, fr, cc.threads, -1,ps,qr);
        qr[1]-=qr[0];
        if qr[1]<simg_mina then qr[1]:=simg_mina;
        qr[1]:=1/qr[1];
    end;
    fw.pt:=nil; rm.pt:=nil;
    SImg_New(cc, fw, fi.nx,fi.ny,1, false);
    SImg_New(cc, rm, SImg_RawByte, fi.nx,fi.ny,1, true);  { interpolation mask }
    for i:=0 to fw.size-1 do
    begin
        r:=(fi.pt[i]-qi[0])*qi[1] - 1;
        if r<0 then r:=1
        else r:=1/(1+sqr(r));
        if fr.pt<>nil then
        begin
            s:=(fr.pt[i]-qr[0])*qr[1];
            if s>1 then
            begin
                r/=1+sqr(s-1);
                rm.ptb[i]:=1; { interpolation mask }
            end;
        end;
        fw.pt[i]:=r;
    end;
//    SImg_WriteImageFile(cc, 'bgcorrect-fw0.fits', fw);

    { get residual }
    fg.pt:=nil; fj.pt:=nil;  fk.pt:=nil;
    if abs(fact)<simg_mina then
    begin
        SImg_ApproxW_int(cc, fi,fw,fg,fr, fg, 4,false,pf);      { does not touch fi }
//        SImg_WriteImageFile(cc, 'bgcorrect-fg.fits', fg);
        r:=pf[4];
        fj.pt:=nil;  fk.pt:=nil;
        SImg_New(cc, fj, fi.nx,fi.ny,1, false);
        SImg_New(cc, fk, fi.nx,fi.ny,1, false);
        for i:=0 to fi.size-1 do
        begin
            s:=fi.pt[i]-fg.pt[i]-r*fr.pt[i];
            fk.pt[i]:=s;
            fj.pt[i]:=abs(s);
        end;
        cc.info(cc.name+': Reference image approximation factor: '+real2str(pf[4])+'. Interpolation and calculation of the background correction.' );
    end else
    begin
        if fr.pt<>nil then
        begin
            SImg_New(cc, fk, fi.nx,fi.ny,1, false);
            r:=1/fact;
            for i:=0 to fi.size-1 do
                fk.pt[i]:=fi.pt[i]-r*fr.pt[i];
        end
        else SImg_Duplicate(cc, fi,fk, true);
        SImg_ApproxW_int(cc, fk,fw,fg,fg, fg, 4,false,nil);     { does not touch fk }
        SImg_New(cc, fj, fi.nx,fi.ny,1, false);
        for i:=0 to fi.size-1 do
        begin
            s:=fk.pt[i]-fg.pt[i];
            fk.pt[i]:=s;
            fj.pt[i]:=abs(s);
        end;
    end;
    if SImg_OverWriteAlert(fo,fi) then SImg_Free(cc,fi);
    if SImg_OverWriteAlert(fo,fr) then SImg_Free(cc,fr);
//    SImg_WriteImageFile(cc, 'bgcorrect-res0.fits', fk);

    { get interpolation mask and new weight }
    r:=sqrt(sqr(frad*0.5)+sqr(fwhm*FWHM_TO_RAD*2));
    SImg_Filter(cc, fj,fj, r);
    SImg_FBGFilter(cc, fj,fl, r*4, true);
    for i:=0 to fj.size-1 do
        fj.pt[i]-=fl.pt[i];
    SImg_Free(cc,fl);
//    SImg_WriteImageFile(cc, 'bgcorrect-fj.fits', fj);
    s:=SImg_GetQuantile(cc, fj,1-ir);
    if s<simg_mina then
    begin
        cc.warning('SImg_ImgBGApprox: Residual to small, input and output seem to be equal');
        for i:=0 to fk.size-1 do
            fk.pt[i]+=fg.pt[i];
        SImg_Free(cc,fw);
        SImg_Free(cc,fg);
        SImg_Free(cc,fj);
        SImg_Move(cc, fk,fo);
        SImg_DecDebugLevel;
        exit;
    end;
    s:=1/s;
    for i:=0 to fw.size-1 do
    begin
        r:=1-sqr(fj.pt[i]*s);
        if r<=0 then
        begin
            rm.ptb[i]:=1;
            r:=0;
        end;
        fw.pt[i]*=r+wreg;
    end;
    SImg_Free(cc,fj);
//    SImg_WriteImageFile(cc, 'bgcorrect-fw1.fits', fw);
//    SImg_WriteImageFile(cc, 'bgcorrect-rm.fits', rm);

    { interpolation }
    SImg_InterpolateMasked(cc, fk,rm, 1);
//    SImg_WriteImageFile(cc, 'bgcorrect-res1.fits', fi);
    SImg_Free(cc, rm);

    { masked filtering }
    for i:=0 to fw.size-1 do
        fk.pt[i]*=fw.pt[i];
    SImg_Filter(cc,fk,fk,frad);
    SImg_Filter(cc,fw,fw,frad);
    for i:=0 to fw.size-1 do
        fk.pt[i]:=fk.pt[i]/fw.pt[i] + fg.pt[i];
//    SImg_WriteImageFile(cc, 'bgcorrect-res2.fits', fi);
    SImg_Free(cc,fw);
    SImg_Free(cc,fg);

    SImg_Move(cc, fk,fo);

    SImg_DecDebugLevel;
end;
