{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{*************************************************************************}
{******* basic SImg_FFTImage processing **********************************}
{*************************************************************************}
{*************************************************************************}

{*************************************************************************}
{******* SImg_Clear ******************************************************}
{*************************************************************************}
procedure SImg_Clear(cc:CSImg_Config; fi:TSImg_FFTImage);                                    { fi(*,*):=0; }
var i  : longint;
    d  : ^fftfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Integrity(fi) then cc.error('SImg_Clear: illegal image');
    d:=fi.pt;
    for i:=0 to fi.nxc*fi.ny*2-1 do
        d[i]:=0;
end;

{*************************************************************************}
{******* SImg_CopyR ******************************************************}
{*************************************************************************}
{ Copy fh(y-yt,x-xt,k) --> fi(y,x). Rotate. Do not clear undefined pixels. }
procedure SImg_CopyR(cc:CSImg_Config; fh:TSImg_FPImage; fi:TSImg_FFTImage; k:longint;xt,yt:longint);
var d           : ^simgfp;
    e,e0,e1     : ^fftfp;
    i,j,nx,ny   : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fh) then cc.error('SImg_CopyR: invalid image: fh');
    if (k<0) or (k>=fh.nc) then cc.error('SImg_CopyR: illegal color of fh');
    if not SImg_Integrity(fi) then cc.error('SImg_CopyR: illegal image: fi');
    xt:=xt mod fi.nxr;
    while xt<0 do
        xt+=fi.nxr;
    yt:=yt mod fi.ny;
    while yt<0 do
        yt+=fi.ny;
    nx:=min(fi.nxr,fh.nx);
    ny:=min(fi.ny,fh.ny);
    for i:=0 to ny-1 do
    begin
        d:=fh.pt + i*fh.sl + k;
        e0:=fi.pt + ((yt+i) mod fi.ny)*fi.nxc*2;
        e :=e0 + xt;
        e1:=e0 + fi.nxr;
        for j:=0 to nx-1 do
        begin
            e^:=d^;
            e+=1;
            if e=e1 then e:=e0;
            d+=fh.nc;
        end;
    end;
end;

{ Copy fh(y   ,x   ,k) --> fi(y,x). Do not clear undefined pixels. }
procedure SImg_Copy(cc:CSImg_Config; fh:TSImg_FPImage; fi:TSImg_FFTImage; k:longint);
begin
    SImg_CopyR(cc,fh,fi,k,0,0);
end;

{ Copy fi(y-yt,x-xt) --> fh(y,x,l). Rotate. Do not clear undefined pixels. }
procedure SImg_CopyR(cc:CSImg_Config; fi:TSImg_FFTImage; fh:TSImg_FPImage; l:longint;xt,yt:longint);
var d           : ^simgfp;
    e,e0,e1     : ^fftfp;
    i,j,nx,ny   : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fh) then cc.error('SImg_CopyR: invalid image: fh');
    if (l<0) or (l>=fh.nc) then cc.error('SImg_CopyR: illegal color of fh');
    if not SImg_Integrity(fi) then cc.error('SImg_CopyR: illegal image: fi');
    xt:=-xt; yt:=-yt;
    xt:=xt mod fi.nxr;
    while xt<0 do
        xt+=fi.nxr;
    yt:=yt mod fi.ny;
    while yt<0 do
        yt+=fi.ny;
    nx:=min(fi.nxr,fh.nx);
    ny:=min(fi.ny,fh.ny);
    for i:=0 to ny-1 do
    begin
        d:=fh.pt + i*fh.sl + l;
        e0:=fi.pt + ((yt+i) mod fi.ny)*fi.nxc*2;
        e :=e0 + xt;
        e1:=e0 + fi.nxr;
        for j:=0 to nx-1 do
        begin
            d^:=e^;
            e+=1;
            if e=e1 then e:=e0;
            d+=fh.nc;
        end;
    end;
end;

{ Copy fi(y,x) --> fh(y,x,l). Rotate. Do not clear undefined pixels. }
procedure SImg_Copy(cc:CSImg_Config; fi:TSImg_FFTImage; fh:TSImg_FPImage; l:longint);
begin
    SImg_CopyR(cc,fi,fh,l,0,0);
end;

{ Copy fi(y,x) --> fh(y,x,l). Clear undefined pixels. }
procedure SImg_CopyC(cc:CSImg_Config; fi:TSImg_FPImage; ti:TSImg_FFTImage; l:longint);
var x,y,i,j  : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Integrity(ti) then cc.error('SImg_CopyC: invalid image: ti');
    if (l<0) or (l>=fi.nc) then cc.error('SImg_CopyC: illegal color of fi');
    if not SImg_Recover(cc,fi) then cc.error('SImg_CopyC: illegal image: fi');
    for y:=0 to ti.ny-1 do
    begin
        j:=y*ti.nxc*2;
        x:=0;
        if y<fi.ny then
        begin
            i:=y*fi.sl+l;
            for x:=0 to fi.nx-1 do
            begin
                ti.pt[x+j]:=fi.pt[i];
                i+=fi.nc;
            end;
            x:=fi.nx;
        end;
        for x:=x to ti.nxr-1 do
            ti.pt[x+j]:=0;
    end;

end;

{*************************************************************************}
{******* SImg_Konj_C *****************************************************}
{*************************************************************************}
procedure SImg_Konj_C(cc:CSImg_Config; fi:TSImg_FFTImage);
var d   : ^fftfp;
    i   : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Integrity(fi) then cc.error('SImg_Konj_C: illegal image: fi');
    d:=@fi.pt[1];
    for i:=0 to fi.nxc*fi.ny-1 do
        d[i*2]:=-d[i*2];
end;

{*************************************************************************}
{******* SImg_Mul_C ******************************************************}
{*************************************************************************}
procedure SImg_Mul_C(cc:CSImg_Config; fi,fj,fk:TSImg_FFTImage);  { fi*fj --> fk, images may be equal }
var d           : dword;
    rr,ri,sr,si : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Integrity(fi) then cc.error('SImg_Mul_C: illegal image: fi');
    if not SImg_Integrity(fj) then cc.error('SImg_Mul_C: illegal image: fj');
    if not SImg_Integrity(fk) then cc.error('SImg_Mul_C: illegal image: fk');
    if (fi.nxc<>fj.nxc) or (fi.ny<>fj.ny) then cc.error('SImg_Mul_C: incompatible images (size): fj');
    if (fi.nxc<>fk.nxc) or (fi.ny<>fk.ny) then cc.error('SImg_Mul_C: incompatible images (size): fk');
    for d:=0 to fi.nxc*fi.ny-1 do
    begin
        rr:=fi.pt[d*2]; ri:=fi.pt[d*2+1];
        sr:=fj.pt[d*2]; si:=fj.pt[d*2+1];
        fk.pt[d*2  ]:=rr*sr-ri*si;
        fk.pt[d*2+1]:=ri*sr+rr*si;
    end;
end;

{*************************************************************************}
{******* SImg_MulK_C *****************************************************}
{*************************************************************************}
procedure SImg_MulK_C(cc:CSImg_Config; fi,fj,fk:TSImg_FFTImage);  {  fi*fj~ --> fk }
var d           : dword;
    rr,ri,sr,si : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Integrity(fi) then cc.error('SImg_MulK_C: illegal image: fi');
    if not SImg_Integrity(fj) then cc.error('SImg_MulK_C: illegal image: fj');
    if not SImg_Integrity(fk) then cc.error('SImg_MulK_C: illegal image: fk');
    if (fi.nxc<>fj.nxc) or (fi.ny<>fj.ny) then cc.error('SImg_MulK_C: incompatible images (size)');
    if (fi.nxc<>fk.nxc) or (fi.ny<>fk.ny) then cc.error('SImg_MulK_C: incompatible images (size): fk');
    for d:=0 to fi.nxc*fi.ny-1 do
    begin
        rr:=fi.pt[d*2]; ri:= fi.pt[d*2+1];
        sr:=fj.pt[d*2]; si:=-fj.pt[d*2+1];
        fk.pt[d*2  ]:=rr*sr-ri*si;
        fk.pt[d*2+1]:=ri*sr+rr*si;
    end;
end;

{*************************************************************************}
{******* SImg_Inv_C ******************************************************}
{*************************************************************************}
procedure SImg_Inv_C(cc:CSImg_Config; fi:TSImg_FFTImage);
var d       : ^fftfp;
    a,b,s   : double;
    i       : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Integrity(fi) then cc.error('SImg_InvC: illegal image: fi');
    d:=fi.pt;
    for i:=0 to fi.nxc*fi.ny-1 do
    begin
        a:=d[0]; b:=d[1];
        s:=sqr(a)+sqr(b);
        if s<simg_mina then s:=simg_mina;
        s:=1/s;
        d[0]:=a*s; d[1]:=-b*s;
        d+=2;
    end;
end;

{*************************************************************************}
{******* SImg_Div_C ******************************************************}
{*************************************************************************}
procedure SImg_Div_C(cc:CSImg_Config; fi,fj,fk:TSImg_FFTImage);       { fi/fj --> fk }
var d,e,f         : ^fftfp;
    rr,ri,sr,si,s : double;
    i             : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Integrity(fi) then cc.error('SImg_Div: illegal image: fi');
    if not SImg_Integrity(fj) then cc.error('SImg_Div: illegal image: fj');
    if not SImg_Integrity(fk) then cc.error('SImg_Div: illegal image: fk');
    if (fi.nxc<>fj.nxc) or (fi.ny<>fj.ny) or (fi.nxc<>fk.nxc) or (fi.ny<>fk.ny) then cc.error('SImg_Div_FCFC: illegal images');
    d:=fi.pt; e:=fj.pt; f:=fk.pt;
    for i:=0 to fi.nxc*fi.ny-1 do
    begin
        rr:=d[0]; ri:=d[1];
        sr:=e[0]; si:=e[1];
        s:=sqr(sr)+sqr(si);
        if s<simg_mina then s:=simg_mina;
        s:=1/s;
        f[0]:=(rr*sr+ri*si)*s;
        f[1]:=(ri*sr-rr*si)*s;
        d+=2; e+=2; f+=2;
    end;
end;

{*************************************************************************}
{******* SImg_Exp_C  *****************************************************}
{*************************************************************************}
procedure SImg_Exp_C(cc:CSImg_Config; fi:TSImg_FFTImage; n:double);
var d           : dword;
    a,b,w,r     : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Integrity(fi) then cc.error('SImg_Exp: illegal FFTImage ');
    for d:=0 to fi.nxc*fi.ny-1 do
    begin
        a:=fi.pt[d*2]; b:=fi.pt[d*2+1];
        if a<simg_mina then a:=simg_mina;
        w:=arctan(b/a);
        if a<0 then w+=pi;
        w*=n;
        r:=sqr(a)+sqr(b);
        if r<simg_mina then r:=simg_mina;
        r:=exp(ln(r)*n*0.5);
        fi.pt[d*2]:=r*cos(w);
        fi.pt[d*2+1]:=r*sin(w);
    end;
end;

{*************************************************************************}
{******* SIMG_FFTWrite ***************************************************}
{*************************************************************************}
{ Write real FFT image to file. Not very efficient and should only be used for debugging.
   fn    : filename
   ti    : real image }
procedure SImg_FFTWrite(cc:CSImg_Config; fn:ansistring; ti:TSImg_FFTImage);
var fi     : TSImg_FPImage;
begin
    if cc=nil then cc:=SImg_Config_Default;
    fi.pt:=nil;
    SImg_New(cc, fi, ti.nxr,ti.ny,1);
    SImg_Copy(cc,ti,fi,0);
    SImg_WriteImageFile(cc,fn,fi);
    SImg_Free(cc,fi);
end;


{*************************************************************************}
{******* SIMG_Convolute **************************************************}
{*************************************************************************}
{ fj = fi \ast fs
   fi   : input
   fs   : PSF
   fj   : output (must be allocated)
   xc,yc: PSF center coordinates
}
procedure SImg_Convolute(cc:CSImg_Config; fi,fs,fj:TSImg_FPImage; xc,yc: longint);
var ti,ts       : TSImg_FFTImage;
    c           : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Convolute: fi='+SImg_InfoStr2(fi)+' fs='+SImg_InfoStr2(fs)+' fj='+SImg_InfoStr2(fj)+' xc='+int2str(xc)+' yc='+int2str(yc));
    if not SImg_Recover(cc,fi) then cc.error('SImg_Convolaute: Invalid image: fi');
    if not SImg_Recover(cc,fj) then cc.error('SImg_Convolaute: Invalid image: fj');
    if fj.nc<>fi.nc then cc.error('SImg_Convolaute: Invalid dimensions: fj');
    if not SImg_Recover(cc,fs) then cc.error('SImg_Convolaute: Invalid image: fpsf');
    if (fs.nc<>1) then cc.error('SImg_Convolaute: Invalid dimensions: fpsf');

    ti.pt:=nil;  ts.pt:=nil;
    SImg_New_findbase(cc, ti, fi.nx+fs.nx-1, fi.ny+fs.ny-1);
    SImg_New(cc, ts, ti.nxr,ti.ny);
    SImg_CopyR(cc, fs,ts, 0, -xc,-yc);
    SImg_FFT_r2c(cc,ts);
    for c:=0 to fi.nc-1 do
    begin
        SImg_CopyC(cc, fi,ti, c);
        SImg_FFT_r2c(cc,ti);
        SImg_Mul_C(cc, ti, ts, ti);
        SImg_FFT_c2r(cc,ti);
        SImg_Copy(cc, ti,fj, c);
    end;
    SImg_Free(cc,ti);
    SImg_Free(cc,ts);
    SImg_DecDebugLevel;
end;
