{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

// define[debug_filter]

{*************************************************************************}
{******* SImg_Filter *****************************************************}
{*************************************************************************}
{******* SImg_Filter_quantile ********************************************}
{ q: quantile }
threadprocedure SImg_Filter_quantile(fi:TSImg_FPImage; fk:TSImg_FPImage; ia:longint; iz:longint; rad:longint; q:single; ss:longint);
var buf,d,e            : ^simgfp;
    r,qf               : simgfp;
    size,xa,ya,xz,yz   : longint;
    i,j,x,y,c,k,l,m,n  : longint;
    x0,y0              : longint;
begin
    if rad<1 then rad:=1;
    size:=sqr(2*rad+1) div 2 + 1;
    getmem2(buf, sizeof(simgfp)*size);
    x0:=((fi.nx-1) mod ss) div 2;
    y0:=((fi.ny-1) mod ss) div 2;
    qf:=1;
    if q>0.5 then
    begin
        qf:=-1;
        q:=1-q;
    end;
    if q<1/16 then q:=1/16;

    for i:=ia to iz do
    begin
        ya:=y0+i*ss-rad; if ya<0 then ya:=0;
        yz:=y0+i*ss+rad; if yz>=fi.ny then yz:=fi.ny-1;
        e:=fk.pt + i*fk.sl;
        for j:=0 to fk.nx-1 do
        begin
            xa:=x0+j*ss-rad; if xa<0 then xa:=0;
            xz:=x0+j*ss+rad; if xz>=fi.nx then xz:=fi.nx-1;
            r:=(yz-ya+1)*(xz-xa+1)*q;
            m:=trunc(r+0.05);
            n:=trunc(r-0.05);
            for c:=0 to fi.nc-1 do
            begin
                k:=0;
                for y:=ya to yz do
                begin
                    d:=fi.pt + y*fi.sl + xa*fi.nc + c;
                    for x:=xa to xz do
                    begin
                        r:=d^*qf; d+=fi.nc;
                        if k<=m then
                        begin
                            buf[k]:=r;
                            k+=1;
                        end
                        else if r<buf[m] then buf[m]:=r;
                        l:=k-1;
                        while (l>0) and (r<buf[l-1]) do  // buf[l]=r
                        begin
                            buf[l]:=buf[l-1];
                            buf[l-1]:=r;
                            l-=1;
                        end;
                    end;
                end;
                e^:=0.5*qf*(buf[n]+buf[m]); { quantile filter }
                e+=1;
            end;
        end;
    end;
    freemem2(buf, sizeof(simgfp)*size);
end;

{******* SImg_Filter_rquantile ********************************************}
{ q: quantile }
threadprocedure SImg_Filter_rquantile(fi:TSImg_FPImage; fk:TSImg_FPImage; ia:longint; iz:longint; rrad:single; q:single; ss:longint);
var buf,d,e            : ^simgfp;
    r,qf               : simgfp;
    size,xa,ya,xz,yz   : longint;
    i,j,x,y,c,k,l,m,n  : longint;
    x0,y0,xc,yc,o,rad  : longint;
    rbuf               : ^longint;
begin
    rrad+=0.01;
    if rrad<1.6 then rrad:=1.6;
    rad:=trunc(sqrt(sqr(rrad)-1.00001));
    size:=sqr(2*rad+1) div 2 + 1;
    getmem(buf, sizeof(simgfp)*size);
    x0:=((fi.nx-1) mod ss) div 2;
    y0:=((fi.ny-1) mod ss) div 2;
    qf:=1;
    if q>0.5 then
    begin
        qf:=-1;
        q:=1-q;
    end;
    if q<1/16 then q:=1/16;

    getmem(rbuf,(2*rad+1)*sizeof(longint));
    m:=0;
    for y:=-rad to rad do
    begin
        x:=trunc(sqrt(sqr(rrad)-sqr(y)-1));
        rbuf[y+rad]:=x;
        if x>=0 then m+=2*x+1;
    end;
    m:=trunc(m*q+0.05);

    for i:=ia to iz do
    begin
        yc:=y0+i*ss;
        ya:=yc-rad; if ya<0 then ya:=0;
        yz:=yc+rad; if yz>=fi.ny then yz:=fi.ny-1;
        e:=fk.pt + i*fk.sl;
        for j:=0 to fk.nx-1 do
        begin
            xc:=x0+j*ss;
            for c:=0 to fi.nc-1 do
            begin
                k:=0; o:=0;
                for y:=ya to yz do
                begin
                    x:=rbuf[y-yc+rad];
                    xa:=xc-x; if xa<0 then xa:=0;
                    xz:=xc+x; if xz>=fi.nx then xz:=fi.nx-1;
                    d:=fi.pt + y*fi.sl + xa*fi.nc + c;
                    for x:=xa to xz do
                    begin
                        o+=1;
                        r:=d^*qf;
                        if k<=m then
                        begin
                            buf[k]:=r;
                            k+=1;
                        end
                        else if r<buf[m] then buf[m]:=r;
                        l:=k-1;
                        while (l>0) and (r<buf[l-1]) do  // buf[l]=r
                        begin
                            buf[l]:=buf[l-1];
                            buf[l-1]:=r;
                            l-=1;
                        end;
                        d+=fi.nc;
                    end;
                end;
                r:=o*q;
                o:=trunc(r+0.05);
                n:=trunc(r-0.05);
                e^:=0.5*qf*(buf[n]+buf[o]); { quantile filter }
                e+=1;
            end;
        end;
    end;
    freemem(buf, sizeof(simgfp)*size);
    freemem(rbuf);
end;

{******* SImg_Filter_avg *************************************************}
threadprocedure SImg_Filter_avg(fi:TSImg_FPImage; fk:TSImg_FPImage; ia:longint; iz:longint; rad:longint; ss:longint);
var d,e           : ^simgfp;
    r,s           : double;
    xa,ya,xz,yz   : longint;
    i,j,x,y,c     : longint;
    x0,y0         : longint;
begin
    x0:=((fi.nx-1) mod ss) div 2;
    y0:=((fi.ny-1) mod ss) div 2;
    if rad<1 then rad:=1;
    for i:=ia to iz do
    begin
        ya:=y0+i*ss-rad; if ya<0 then ya:=0;
        yz:=y0+i*ss+rad; if yz>=fi.ny then yz:=fi.ny-1;
        e:=fk.pt + i*fk.sl;
        for j:=0 to fk.nx-1 do
        begin
            xa:=x0+j*ss-rad; if xa<0 then xa:=0;
            xz:=x0+j*ss+rad; if xz>=fi.nx then xz:=fi.nx-1;
            s:=1/((yz-ya+1)*(xz-xa+1));
            for c:=0 to fi.nc-1 do
            begin
                r:=0;
                for y:=ya to yz do
                begin
                    d:=fi.pt + y*fi.sl + xa*fi.nc + c;
                    for x:=xa to xz do
                    begin
                        r+=d^; d+=fi.nc;
                    end;
                end;
                e^:=r*s;
                e+=1;
            end;
        end;
    end;
end;

{******* SImg_Filter_ravg *************************************************}
threadprocedure SImg_Filter_ravg(fi:TSImg_FPImage; fk:TSImg_FPImage; ia:longint; iz:longint; rrad:single; ss:longint);
var d,e           : ^simgfp;
    r             : double;
    xa,ya,xz,yz   : longint;
    i,j,x,y,c     : longint;
    x0,y0         : longint;
    xc,yc,rad,n   : longint;
    rbuf          : ^longint;
begin
    rrad+=0.01;
    if rrad<1.6 then rrad:=1.6;
    rad:=trunc(sqrt(sqr(rrad)-1.00001));
    x0:=((fi.nx-1) mod ss) div 2;
    y0:=((fi.ny-1) mod ss) div 2;

    getmem(rbuf,(2*rad+1)*sizeof(longint));
    for y:=-rad to rad do
        rbuf[y+rad]:=trunc(sqrt(sqr(rrad)-sqr(y)-1));

    for i:=ia to iz do
    begin
        yc:=y0+i*ss;
        ya:=yc-rad; if ya<0 then ya:=0;
        yz:=yc+rad; if yz>=fi.ny then yz:=fi.ny-1;
        e:=fk.pt + i*fk.sl;
        for j:=0 to fk.nx-1 do
        begin
            xc:=x0+j*ss;
            for c:=0 to fi.nc-1 do
            begin
                r:=0; n:=0;
                for y:=ya to yz do
                begin
                    x:=rbuf[y-yc+rad];
                    xa:=xc-x; if xa<0 then xa:=0;
                    xz:=xc+x; if xz>=fi.nx then xz:=fi.nx-1;
                    d:=fi.pt + y*fi.sl + xa*fi.nc + c;
                    for x:=xa to xz do
                    begin
                        r+=d^;
                        n+=1;
                        d+=fi.nc;
                    end;
                end;
                e^:=r/n;
                e+=1;
            end;
        end;
    end;
    freemem(rbuf);
end;

{******* SImg_Filter_rravg *************************************************}
{ 1/( 1/f1 + 1/f2 + ... ) }
threadprocedure SImg_Filter_rravg(fi:TSImg_FPImage; fk:TSImg_FPImage; ia:longint; iz:longint; rrad:single; ss:longint);
var d,e           : ^simgfp;
    r             : double;
    s             : simgfp;
    xa,ya,xz,yz   : longint;
    i,j,x,y,c     : longint;
    x0,y0         : longint;
    xc,yc,rad,n   : longint;
    rbuf          : ^longint;
begin
    rrad+=0.01;
    if rrad<1.6 then rrad:=1.6;
    rad:=trunc(sqrt(sqr(rrad)-1.00001));
    x0:=((fi.nx-1) mod ss) div 2;
    y0:=((fi.ny-1) mod ss) div 2;

    getmem(rbuf,(2*rad+1)*sizeof(longint));
    for y:=-rad to rad do
        rbuf[y+rad]:=trunc(sqrt(sqr(rrad)-sqr(y)-1));

    for i:=ia to iz do
    begin
        yc:=y0+i*ss;
        ya:=yc-rad; if ya<0 then ya:=0;
        yz:=yc+rad; if yz>=fi.ny then yz:=fi.ny-1;
        e:=fk.pt + i*fk.sl;
        for j:=0 to fk.nx-1 do
        begin
            xc:=x0+j*ss;
            xa:=xc-rad; if xa<0 then xa:=0;
            xz:=xc+rad; if xz>=fi.nx then xz:=fi.nx-1;
            for c:=0 to fi.nc-1 do
            begin
                r:=0; n:=0;
                for y:=ya to yz do
                begin
                    x:=rbuf[y-yc+rad];
                    xa:=xc-x; if xa<0 then xa:=0;
                    xz:=xc+x; if xz>=fi.nx then xz:=fi.nx-1;
                    d:=fi.pt + y*fi.sl + xa*fi.nc + c;
                    for x:=xa to xz do
                    begin
                        s:=d^;
                        if s<simg_mina then s:=simg_mina;
                        r+=1/s;
                        n+=1;
                        d+=fi.nc;
                    end;
                end;
                e^:=n/r;
                e+=1;
            end;
        end;
    end;
    freemem(rbuf);
end;

{******* SImg_Filter_gauss ***********************************************}
threadprocedure SImg_Filter_gauss(fi:TSImg_FPImage; fk:TSImg_FPImage; ia:longint; iz:longint; rad:double; ss:longint);
var d,e,f,psf   : ^simgfp;
    r,s,t       : double;
    fdia,frad   : longint;  { filter radius / diameter}
    xa,ya,xz,yz : longint;
    i,j,x,y,c   : longint;
    x0,y0,xs,ys : longint;
begin
    x0:=((fi.nx-1) mod ss) div 2;
    y0:=((fi.ny-1) mod ss) div 2;
    if rad<0.1 then rad:=0.1;
    frad:=round(rad*2+0.49);
    fdia:=frad*2+1;
    getmem2(psf,sizeof(simgfp)*sqr(fdia));

    r:=-1/sqr(rad);
    s:=exp(sqr(frad+1)*r);
    d:=psf;
    for y:=-frad to frad do
        for x:=-frad to frad do
        begin
            t:=exp((sqr(x)+sqr(y))*r)-s;
            if t<0 then t:=0;
            d^:=t;
            d+=1;
        end;

    for i:=ia to iz do
    begin
        ys:=y0+i*ss;
        ya:=ys-frad; if ya<0 then ya:=0;
        yz:=ys+frad; if yz>=fi.ny then yz:=fi.ny-1;
        e:=fk.pt + i*fk.sl;
        for j:=0 to fk.nx-1 do
        begin
            xs:=x0+j*ss;
            xa:=xs-frad; if xa<0 then xa:=0;
            xz:=xs+frad; if xz>=fi.nx then xz:=fi.nx-1;
            s:=0;
            for c:=0 to fk.nc-1 do
                e[c]:=0;
            for y:=ya to yz do
            begin
                d:=fi.pt + y*fi.sl + xa*fi.nc;
                f:=psf + (y-ys+frad)*fdia + (xa-xs+frad);
                for x:=xa to xz do
                begin
                    r:=f^; f+=1;
                    s+=r;
                    for c:=0 to fi.nc-1 do
                        e[c]+=d[c]*r;
                    d+=fi.nc;
                end;
            end;
            s:=1/s;
            for c:=0 to fk.nc-1 do
                e[c]*=s;
            e+=fk.nc;
        end
    end;
    freemem2(psf,sizeof(simgfp)*sqr(fdia));
end;

{******* SImg_Filter_drop ************************************************}
threadprocedure SImg_Filter_drop(fi:TSImg_FPImage; fk:TSImg_FPImage; ia:longint; iz:longint; rad:longint; dd:boolean; db:boolean);
var d,e,f         : ^simgfp;
    r,s           : simgfp;
    bd,bb         : boolean;
    x1,x2,y1,y2   : longint;
    xa,ya,xz,yz   : longint;
    i,j,x,y,c     : longint;
begin
    if rad<1 then rad:=1;
    for i:=ia to iz do
    begin
        ya:=i-rad; if ya<0 then ya:=0;
        yz:=i+rad; if yz>=fi.ny then yz:=fi.ny-1;
        if i>0 then y1:=-fi.sl
        else y1:=fi.sl;
        if i<fi.ny-1 then y2:=fi.sl
        else y2:=-fi.sl;
        e:=fk.pt + i*fk.sl;
        f:=fi.pt + i*fi.sl;
        for j:=0 to fi.nx-1 do
        begin
            xa:=j-rad; if xa<0 then xa:=0;
            xz:=j+rad; if xz>=fi.nx then xz:=fi.nx-1;
            if j>0 then x1:=-fi.nc
            else x1:=fi.nc;
            if j<fi.nx-1 then x2:=fi.nc
            else x2:=-fi.nc;
            for c:=0 to fi.nc-1 do
            begin
                s:=f^;
                bd:=dd; bb:=db;
                for y:=ya to yz do
                begin
                    d:=fi.pt + y*fi.sl + xa*fi.nc + c;
                    for x:=xa to xz do
                    begin
                        r:=d^; d+=fi.nc;
                        bd:=bd and (s<=r);
                        bb:=bb and (s>=r);
                    end;
                end;
                if bd or bb then e^:=0.25*(f[x1]+f[x2]+f[y1]+f[y2])
                else e^:=f^;
                e+=1; f+=1;
            end;
        end;
    end;
end;

{******* SImg_FilterGauss_thread *****************************************}
threadprocedure SImg_FilterGauss_thread(cc:CSImg_Config; fi:TSImg_FPImage; fk:TSImg_FPImage; ya:longint; yz:longint; rad:double; frad:TSImg_RDImage; frad_limit:longint);
var d,e             : ^simgfp;
    b               : ^byte;
    s1,s2,w1,w2,rw  : double;
    by,fradd        : boolean;
    r1,r2,b1,b2     : longint;
    x,y,c,k1,k2     : longint;
    dx1,dx2,dy1,dy2 : longint;

begin
    { sqr(sig) = 2/3*sqr(rad) }
    if rad<0.0011 then rad:=0.0011;
    r1:=round(rad+0.499);
    r2:=round(sqrt(0.5)*rad+0.499);
    w1:=sqr(rad)/(3*sqr(r1)-2*sqr(rad));
    w2:=sqr(rad)/(6*sqr(r2)-2*sqr(rad));
//ifdef[debug_filter]
    if ya=0 then cc.info('SImg_FilterGauss: rad='+real2str(rad)+' frad_limit='+int2str(frad_limit)+' r1='+int2str(r1)+' r2='+int2str(r2)+' w1='+real2str(w1)+' w2='+real2str(w2));
//endif
    dx1:=r1*fi.nc;
    dy1:=r1*fi.sl;
    dx2:=r2*fi.nc;
    dy2:=r2*fi.sl;
    d:=fi.pt + ya*fi.sl;
    e:=fk.pt + ya*fk.sl;
    rw:=1/(1+4*w1+4*w2);
    b:=frad.ptb + ya*frad.sl;
    b1:=1;
    b2:=0;
    fradd:=frad.pt<>nil;
    if frad.size<>fi.size then
    begin
        b1:=0;
        b2:=frad.nc;
    end;
    for y:=ya to yz do
    begin
        by:=(y>=r1) and (y+r1<fi.ny);
        for x:=0 to fi.nx-1 do
        begin
            if by and (x>=r1) and (x+r1<fi.nx) then
                for c:=0 to fi.nc-1 do
                begin
                    if fradd and (b^>frad_limit) then e^:=d^
                    else e^:=( w2*(d[-dx2-dy2]+d[dx2-dy2]+d[-dx2+dy2]+d[dx2+dy2]) + w1*(d[-dx1]+d[dx1]+d[-dy1]+d[dy1]) + d^ )*rw;
                    d+=1; e+=1; b+=b1;
                end
            else
                for c:=0 to fi.nc-1 do
                begin
                    if fradd and (b^>frad_limit) then e^:=d^
                    else begin
                        s1:=0; s2:=0; k1:=0; k2:=0;
                        if (x+r2<fi.nx) and (y+r2<fi.ny) then begin s2:=d[ dx2+dy2]; k2+=1; end;
                        if (x>=r2) and (y+r2<fi.ny)      then begin s2+=d[-dx2+dy2]; k2+=1; end;
                        if (x+r2<fi.nx) and (y>=r2)      then begin s2+=d[ dx2-dy2]; k2+=1; end;
                        if (x>=r2) and (y>=r2)           then begin s2+=d[-dx2-dy2]; k2+=1; end;
                        if (x+r1<fi.nx) then begin s1+=d[ dx1]; k1+=1; end;
                        if (x>=r1)      then begin s1+=d[-dx1]; k1+=1; end;
                        if (y+r1<fi.ny) then begin s1+=d[ dy1]; k1+=1; end;
                        if (y>=r1)      then begin s1+=d[-dy1]; k1+=1; end;
                        e^:=(d^ + w1*s1 + w2*s2) / (1 + k1*w1 + k2*w2);
                    end;
                    d+=1; e+=1; b+=b1;
                end;
            b+=b2;
        end;
    end;
end;

{******* SImg_FilterGauss ************************************************}
{ Gauss filter: exp( -sqr(r/rad)) = exp( -sqr(r)/(2*sqr(sig)) ) where sig=rad/sqrt(2)
  for internal use only
  fi    : input
  fj    : output
  sig   : sigma
}
procedure SImg_FilterGauss(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; rad:double);
var frad          : TSImg_RDImage;  { just a dummy }
    threads       : PSImg_FilterGauss_thread;
    fk1,fk2,fk3   : TSImg_FPImage;
    ss,rr,rf,r    : double;
    i,tn,ya,yz    : longint;
    j,jz          : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_FilterGauss: Invalid image: fi');
    tn:=cc.threads;
    getmem(threads,tn*sizeof(CSImg_FilterGauss_thread));
    for i:=0 to tn-1 do
        threads[i]:=CSImg_FilterGauss_thread.create;
    fk1.pt:=nil;
    fk2.pt:=nil;
    if SImg_OverwriteAlert(fi,fj)
    then fk1:=fi
    else SImg_Duplicate(cc,fi,fk1,true);
    SImg_Duplicate(cc,fi,fk2,false);

    { calculate last radius }
    rf:=1;
//    ss:=1.8*sqr(sig);   // empirical; theoretical: ss:=1.5*sqr(sig)
    ss:=0.9*sqr(rad);   // empirical; theoretical: ss:=0.75*sqr(rad)
//ifdef[debug_filter]
    cc.info('SImg_FilterGauss: ss='+real2str(ss));
//endif
    if ss<3 then        // ensure at least 3 iterations
    begin
        rf:=sqrt(ss/(6-ss));    // empirical; theoretical: rs:=sqrt(ss/3)
        ss:=3;
    end;
    rr:=0.95;
    jz:=0;
    while ss>sqr(round(rr))+0.25 do
    begin
        ss-=sqr(round(rr));
        rr/=0.8;
        jz+=1;
    end;
    r:=sqrt(ss);

    { smoothing }
    frad.pt:=nil;
    for j:=0 to jz do
    begin
        ya:=0;
        for i:=0 to tn-1 do
        begin
            yz:=fi.ny*(i+1) div tn;
            threads[i].setwork(cc,fk1,fk2, ya,yz-1, r*rf, frad,0);
            ya:=yz;
        end;
        for i:=0 to tn-1 do
            threads[i].wait;
        fk3:=fk1;
        fk1:=fk2;
        fk2:=fk3;
        rr*=0.8;
        r:=round(rr);
    end;

    for i:=0 to tn-1 do
        threads[i].destroy;
    freemem(threads);
    SImg_Free(cc,fk2);
    SImg_Move(cc,fk1,fj);
end;

{******* SImg_Filter *****************************************************}
{ Gauss filter with variable radius : exp( -sqr(r/(rad*frad)) )
  The radius is controlled by iterations. This is not very exact but fast.
  fi    : input
  fj    : output
  rad   : maximum filter radius
  frad  : radius factor: 0 means 1.0, 255 means 0.0, frad.size=fi.size or frad.nx*frad.ny=fi.nx*fi.ny
}
procedure SImg_Filter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; rad:double; frad:TSImg_RDImage);
var threads       : PSImg_FilterGauss_thread;
    fk1,fk2,fk3   : TSImg_FPImage;
    ss,rr,rf,r    : double;
    ss0,ss1       : double;
    i,tn,ya,yz    : longint;
    j,jz          : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Filter: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj)+' rad='+real2str(rad)+' frad='+SImg_InfoStr2(frad));

    if not SImg_Recover(cc,fi) then cc.error('SImg_Filter: Invalid image: fi');
    if not SImg_Recover(cc,frad) then cc.error('SImg_Filter: Invalid image: frad');
    if frad.fmt<>SImg_RawByte then cc.error('SImg_Filter: Invalid data format: frad');
    if (frad.nx<>fi.nx) or (frad.ny<>fi.ny) or ((frad.nc<>fi.nc) and (frad.nc<>1)) then cc.error('SImg_Filter: Invalid size: frad');

    tn:=cc.threads;
    getmem(threads,tn*sizeof(CSImg_FilterGauss_thread));
    for i:=0 to tn-1 do
        threads[i]:=CSImg_FilterGauss_thread.create;
    fk1.pt:=nil;
    fk2.pt:=nil;
    if SImg_OverwriteAlert(fi,fj)
    then fk1:=fi
    else SImg_Duplicate(cc,fi,fk1,true);
    SImg_Duplicate(cc,fi,fk2,false);

    { calculate last radius }
    rf:=1;
//    ss:=1.8*sqr(sig);   // empirical; theoretical: ss:=1.5*sqr(sig)
    ss:=0.9*sqr(rad);   // empirical; theoretical: ss:=0.75*sqr(rad)
//ifdef[debug_filter]
    cc.info('SImg_FilterGauss: ss='+real2str(ss));
//endif
    if ss<5 then        // ensure at least 3 iterations
    begin
        i:=round(ss*4);
        if i<1 then i:=1;
        if i>5 then i:=5;
        rf:=sqrt(ss/i);
        ss:=i;
    end;
    ss0:=ss;
    rr:=0.6;
    jz:=0;
    while ss>sqr(round(rr))+0.25 do
    begin
        ss-=sqr(round(rr));
        rr/=0.8;
        jz+=1;
    end;
    r:=sqrt(ss);

    { smothing }
    ss:=ss0;
    for j:=0 to jz do
    begin
        ss1:=ss-sqr(r);
//ifdef[debug_filter]
        cc.info('SImg_FilterGauss: ss='+real2str(ss) + ' ss1='+real2str(ss1) + ' ss0='+real2str(ss0) + ' frad_limit='+int2str( 255-round(255*sqrt((ss1+ss)/(2*ss0))) ) );
//endif
        ya:=0;
        for i:=0 to tn-1 do
        begin
            yz:=fi.ny*(i+1) div tn;
            threads[i].setwork(cc,fk1,fk2, ya,yz-1, r*rf, frad, 255-round(255*sqrt((ss1+ss)/(2*ss0))) );
            ya:=yz;
        end;
        for i:=0 to tn-1 do
            threads[i].wait;
        fk3:=fk1;
        fk1:=fk2;
        fk2:=fk3;
        rr*=0.8;
        r:=round(rr);
        ss:=ss1;
    end;

    for i:=0 to tn-1 do
        threads[i].destroy;
    freemem(threads);
    SImg_Free(cc,fk2);
    SImg_Move(cc,fk1,fj);
    SImg_DecDebugLevel;
end;


{******* SImg_Filter *****************************************************}
{ fi    : input
  fj    : output
  rad   : filter radius
  algo  : -5 : reciprocal average ( 1/(1/x1+1/x2+...)) ) of round region
          -4 : quantile of round region
          -3 : average of round region
          -2 : quantile of square region
          -1 : average of square region
           0 : Gaussian: exp(-sqr(r/rad))
           1 : drop min of square region
           2 : drop max of square region
           3 : drop min and max of square region
  ss    : step size
  q     : percentage for quantile filter
}
procedure SImg_Filter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; rad:double; algo:longint; ss:longint; q:double);
var fk          : TSImg_FPImage;
    i,tn,ya,yz  : longint;
    irad        : longint;
threadvars;
begin
    fk.pt:=nil;
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Filter: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj)+ ' rad='+real2str(rad)+' algo='+int2str(algo)+' ss='+int2str(ss)+' q='+real2str(q));
    if (ss<1) or (algo>0) then ss:=1;
    if not SImg_Recover(cc,fi) then cc.error('SImg_Filter: Invalid image: fi');
    if (algo<-5) or (algo>3) then cc.error('SImg_Fitler: Invalid algo');
    if not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);

    if (ss=1) and (algo=0) and (rad>1.5) then
    begin
        SImg_FilterGauss(cc,fi,fj, rad);
        SImg_DecDebugLevel;
        exit;
    end;

    SImg_New(cc,fk, ((fi.nx-1) div ss)+1, ((fi.ny-1) div ss)+1, fi.nc);
    ya:=0;
    tn:=cc.threads;
    irad:=round(rad-0.45);
    for i:=0 to tn-1 do
    begin
        yz:=fk.ny*(i+1) div tn;
        case algo of
            -5 : dothread SImg_Filter_rravg(fi,fk, ya,yz-1, rad, ss );
            -4 : dothread SImg_Filter_rquantile(fi,fk, ya,yz-1, rad, q, ss);
            -3 : dothread SImg_Filter_ravg(fi,fk, ya,yz-1, rad, ss );
            -2 : dothread SImg_Filter_quantile(fi,fk, ya,yz-1, irad, q, ss);
            -1 : dothread SImg_Filter_avg(fi,fk, ya,yz-1, irad, ss );
            0  : dothread SImg_Filter_gauss(fi,fk, ya,yz-1, rad, ss );
            else     dothread SImg_Filter_drop(fi,fk, ya,yz-1, irad, (algo and 1)<>0, (algo and 2)<>0 );
        end;
        ya:=yz;
    end;
    threadwait;

    SImg_Move(cc,fk,fj);
    SImg_DecDebugLevel;
end;

{ q=0.5 }
procedure SImg_Filter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; rad:double; algo:longint; ss:longint);
begin
    SImg_Filter(cc, fi,fj, rad, algo, ss, 0.5);
end;

{ Gaussian filter: exp(-sqr(r/rad)) }
procedure SImg_Filter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; rad:double);
begin
    SImg_Filter(cc, fi,fj, rad, 0, 1, 0.5);
end;


{*************************************************************************}
{******* SImg_BGFilter ***************************************************}
{*************************************************************************}
{******* SImg_FBGFilter_thread *******************************************}
threadprocedure SImg_FBGFilter_thread(cc:CSImg_Config; fi:TSImg_FPImage; fk:TSImg_FPImage; mask:TSImg_RDImage; ya:longint; yz:longint; rx:longint; ry:longint; bg:boolean);
var d,e                 : ^simgfp;
    b                   : ^byte;
    r,s                 : simgfp;
    bx2,bx3,by2,by3     : boolean;
    x,y,c,rx2,ry2       : longint;
    dx1,dx2,dy1,dy2     : longint;
    dby1,dby2           : longint;
begin
    if rx<1 then rx:=1;
    if rx>fi.nx div 2 then rx:=fi.nx div 2;
    rx2:=round(sqrt(0.5)*rx);
    if ry<1 then ry:=1;
    if ry>fi.ny div 2 then ry:=fi.ny div 2;
    ry2:=round(sqrt(0.5)*ry);
    dx1:=rx*fi.nc;
    dy1:=ry*fi.sl;
    dx2:=rx2*fi.nc;
    dy2:=ry2*fi.sl;
    dby1:=ry*mask.sl;
    dby2:=ry2*mask.sl;
    d:=fi.pt + ya*fi.sl;
    e:=fk.pt + ya*fk.sl;
    if mask.ptb=nil then b:=nil
    else b:=mask.ptb + ya*mask.sl;
{ unmasked core }
//define[core_u(][)][
    for y:=ya to yz do
    begin
        by2:=(y>=ry) and (y+ry<fi.ny);
        by3:=(rx2>0) and (ry2>0) and (y>=ry2) and (y+ry2<fi.ny);
        for x:=0 to fi.nx-1 do
        begin
            bx2:=(x>=rx) and (x+rx<fi.nx);
            bx3:=by3 and (x>=rx2) and (x+rx2<fi.nx);
            for c:=0 to fi.nc-1 do
            begin
                r:=d^;
                if by2 then
                begin
                    s:=0.5*(d[-dy1]+d[dy1]);
                    if s #0 r then r:=s;
                end;
                if bx2 then
                begin
                    s:=0.5*(d[-dx1]+d[dx1]);
                    if s #0 r then r:=s;
                end;
                if bx3 then
                begin
                    s:=0.5*(d[-dx2-dy2]+d[+dx2+dy2]);
                    if s #0 r then r:=s;
                    s:=0.5*(d[-dx2+dy2]+d[+dx2-dy2]);
                    if s #0 r then r:=s;
                end;

                e^:=r;
                d+=1; e+=1;
            end;
        end;
    end]
{ masked core }
//define[core_m(][)][
    for y:=ya to yz do
    begin
        by2:=(y>=ry) and (y+ry<fi.ny) and (b[-dby1]<>0) and (b[dby1]<>0);
        by3:=(rx2>0) and (ry2>0) and (y>=ry2) and (y+ry2<fi.ny);
        for x:=0 to fi.nx-1 do
        begin
            if b^<>0 then
            begin
                bx2:=(x>=rx) and (x+rx<fi.nx) and (b[-rx]<>0) and (b[rx]<>0);
                bx3:=by3 and (x>=rx2) and (x+rx2<fi.nx);
                for c:=0 to fi.nc-1 do
                begin
                    r:=d[c];
                    if by2 then
                    begin
                        s:=0.5*(d[-dy1]+d[dy1]);
                        if s #0 r then r:=s;
                    end;

                    if bx2 then
                    begin
                        s:=0.5*(d[-dx1]+d[dx1]);
                        if s #0 r then r:=s;
                    end;

                    if bx3 then
                    begin
                        if (b[-rx2-dby2]<>0) and (b[rx2+dby2]<>0) then
                        begin
                            s:=0.5*(d[-dx2-dy2]+d[+dx2+dy2]);
                            if s #0 r then r:=s;
                        end;
                        if (b[-rx2+dby2]<>0) and (b[+rx2-dby2]<>0) then
                        begin
                            s:=0.5*(d[-dx2+dy2]+d[+dx2-dy2]);
                            if s #0 r then r:=s;
                        end;
                    end;
                    e[c]:=r;
                end;
            end;
            d+=fi.nc;
            e+=fk.nc;
            b+=1;
        end;
    end]
    if (b=nil) and bg then core_u(<)
    else if bg then core_m(<)
    else if b=nil then core_u(>)
    else core_m(>);
//udefine[core_m(]
//udefine[core_u(]
end;

{******* SImg_FBGFilter **************************************************}
{ Foreground / background approximation filter
  fi    : input
  fj    : output
  mask  : mask, one byte per pixel, 0 means masked out, optional
  rx,ry : approximation radius
  cs    : corner approximation factor, feasible 0..1
  bg    : true for background
  rerun : input is already filtered
  ci    : ignore corners. The size is defined by cs
}
procedure SImg_FBGFilter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; mask:TSImg_RDImage; rx,ry,cs:double; bg,rerun,ci:boolean);
const iter_normal=4;
      iter_rerun=2;
var threads             : PSImg_FBGFilter_thread;
    fk1,fk2,fk3         : TSImg_FPImage;
    i,j,c,x,y,tn        : longint;
    ya,yz,rx1,ry1       : longint;
    iter                : longint;
    rxr,ryr,r           : double;
    d1,d2,d3            : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_FBGFilter: fi='+SImg_InfoStr2(fi)+' mask='+SImg_InfoStr2(mask)+' fj='+SImg_InfoStr2(fj)+' rx='+real2str(rx)+' ry='+real2str(ry)+' cs='+real2str(cs)+' bg='+boolean2str(bg)+' rerun='+boolean2str(rerun)+' ci='+boolean2str(ci));
    if not SImg_Recover(cc,fi) then cc.error('SImg_FBGFilter: Invalid image: fi');
    if mask.pt<>nil then
    begin
        if not SImg_Recover(cc,mask) then cc.error('SImg_FBGFilter: Invalid mask');
        if mask.fmt<>SImg_RawByte then cc.error('SImg_FBGFilter: Invalid data format of mask');
        if (mask.nx<>fi.nx) and (mask.ny<>fi.ny) and (mask.nc<>1) then cc.error('SImg_FBGFilter: Invalid dimensions of mask');
    end;
    if rerun then iter:=iter_rerun
    else iter:=iter_normal;

    tn:=cc.threads;
    getmem(threads,tn*sizeof(CSImg_FBGFilter_thread));
    for i:=0 to tn-1 do
        threads[i]:=CSImg_FBGFilter_thread.create;
    fk1.pt:=nil;
    fk2.pt:=nil;
    if SImg_OverwriteAlert(fi,fj) then fk1:=fi
    else begin
        SImg_Free(cc,fj);
        SImg_Duplicate(cc,fi,fk1,true);
    end;

    { approximate corners }
    if rx<0.1 then rx:=0.1;
    if ry<0.1 then ry:=0.1;
    if cs>1e-8 then
    begin
        rx1:=round(rx*cs);
        if rx1>=fk1.nx then rx1:=fk1.nx-1;
        ry1:=round(ry*cs);
        if ry1>=fk1.ny then ry1:=fk1.ny-1;
        rxr:=1/(cs*rx);
        ryr:=1/(cs*ry);
        for i:=0 to 1 do
            for j:=0 to 1 do
            begin
                d1:=fk1.pt + i*(fk1.ny-1)*fk1.sl + j*(fk1.nx-1)*fk1.nc;
                for y:=0 to ry1 do
                    for x:=0 to rx1 do
                    begin
                        r:=sqr(x*rxr)+sqr(y*ryr);
                        if r>0.999999 then
                        begin
                            d2:=d1 + (1-2*i)*(ry1-y)*fk1.sl + (1-2*j)*(rx1-x)*fk1.nc;
                            if ci then
                            begin
                                r:=1/r;
                                d3:=d1 + (1-2*i)*(ry1-round(y*r))*fk1.sl + (1-2*j)*(rx1-round(x*r))*fk1.nc;
                                for c:=0 to fk1.nc-1 do
                                    d2[c]:=d3[c];
                            end;
                            for c:=0 to fk1.nc-1 do
                                if d2[c]<d1[c] then d1[c]:=d2[c];
                        end;
                    end;
            end;
    end;

    { filtering }
    SImg_Duplicate(cc,fk1,fk2,false);
    for j:=0 to iter-1 do
    begin
        rx1:=2*round(rx/sqrt(iter) + (2*j+1-iter)/(2*iter) );
        ry1:=2*round(ry/sqrt(iter) + (2*j+1-iter)/(2*iter) );
        if (rx1<1) and (ry1<1) then continue;
        repeat
            rx1:=(rx1+1) shr 1;
            ry1:=(ry1+1) shr 1;
//          cc.info('rx1='+real2str(rx1)+' ry1='+real2str(ry1));
            ya:=0;
            for i:=0 to tn-1 do
            begin
                yz:=fi.ny*(i+1) div tn;
                threads[i].setwork(cc,fk1,fk2, mask, ya,yz-1, rx1,ry1, bg);
                ya:=yz;
            end;
            for i:=0 to tn-1 do
                threads[i].wait;
            fk3:=fk1;
            fk1:=fk2;
            fk2:=fk3;
        until (rx1<=1) and (ry1<=1);
    end;

    for i:=0 to tn-1 do
        threads[i].destroy;
    freemem(threads);
    SImg_Free(cc,fk2);
    SImg_Move(cc,fk1,fj);

    SImg_DecDebugLevel;
end;

procedure SImg_FBGFilter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; rad:double; bg:boolean);
var mask : TSImg_RDImage;
begin
    mask.pt:=nil;
    SImg_FBGFilter(cc, fi,fj,mask, rad,rad,0.25, bg, false, false);
end;

procedure SImg_FBGFilter(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; rad:double; bg,rerun:boolean);
var mask : TSImg_RDImage;
begin
    mask.pt:=nil;
    SImg_FBGFilter(cc, fi,fj,mask, rad,rad,0.25, bg, rerun, false);
end;


{*************************************************************************}
{******* SImg_FBGApprox **************************************************}
{*************************************************************************}
{ background approximation using SImg_FBGFilter (local minimum/maximum filter)
    rad   : approximation radii (geometric average of x and y values)
    xyr   : x/y ratio of approximation size.
    cs    : relative approximation size in corner (feasible: 0..1)
    sf    : subtraction factors. If unset and length of rad is 1, the approximation is returned. Otherwise length must be equal to length of rad.
    fr    : radius for quantile pre-filter (<=0 to disable this feature)
    fq    : quantile of pre-filter. <=0 for Gaussian, 0.5 for median
    fg    : approximate foreground instead of background
    ci    : ignore corners. The size is defined by cs
}
procedure SImg_FBGApprox(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; mask:TSImg_RDImage; rad:array of double; xyr,cs:double; sf:array of double; fr,fq:double; fg,ci:boolean);
var fk,fl,fm    : TSImg_FPImage;
    i,j,irad,ss : longint;
    r           : simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BgApprox: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj)+' mask='+SImg_InfoStr2(mask)+' rad.size='+int2str(high(rad)+1)+' xyr='+real2str(xyr,-1)+' cs='+real2str(cs,-1)
        +' sf.size='+int2str(high(sf)+1)+' fr='+real2str(fr,-1)+' fq='+real2str(fq,-1)+' fg='+boolean2str(fg)+' ci='+boolean2str(ci));
    if not SImg_Recover(cc,fi) then cc.error('SImg_BgApprox: Invalid image: fi');
    if mask.pt<>nil then
    begin
        if not SImg_Recover(cc,mask) then cc.error('SImg_FBGFilter: Invalid mask');
        if mask.fmt<>SImg_RawByte then cc.error('SImg_FBGFilter: Invalid data format of mask');
        if (mask.nx<>fi.nx) and (mask.ny<>fi.ny) and (mask.nc<>1) then cc.error('SImg_FBGFilter: Invalid dimensions of mask');
    end;
    if not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);
    if (high(sf)<>high(rad)) and ((high(sf)>=0) or (high(rad)<>0)) then cc.error('SImg_FBGFilter: Number of radii must be equal to number of factors');
    xyr:=sqrt(abs(xyr));
    if xyr<1e-3 then xyr:=1e-3;

    { pre filter }
    fk.pt:=nil;
    if fq<1e-3 then { Gaussian filter }
    begin
        irad:=round(fr+0.3);
        if irad<0 then irad:=0;
        ss:=(irad+1) div 3;
    end
    else begin  { quantile filter }
        irad:=round(fr-0.45);
        if irad<0 then irad:=0;
        ss:=(irad*2+1) div 3;
    end;
    for j:=0 to high(rad) do
    begin
        if xyr>1 then i:=round(rad[j]/xyr*0.25)
        else i:=round(rad[j]*xyr*0.25);
        if i<ss then ss:=i;
    end;
    if ss<1 then ss:=1;
    if irad<=0 then SImg_Duplicate(cc,fi,fk,true)                       { no filter }
    else if fq>=1e-3 then SImg_Filter(cc,fi,fk,irad+0.5,-2,ss, fq)      { quantile filter }
    else if ss<=1 then SImg_Filter(cc, fi,fk, fr)                       { Gaussian filter }
    else begin                                                          { Gaussian filter }
        SImg_Resize(cc, fi,fk, ss,ss, -1);
        SImg_Filter(cc, fk,fk, fr/ss);
    end;


    if high(rad)=0 then
    begin
        { approximations }
        SImg_FBGFilter(cc, fk,fk, mask, rad[0]*xyr/ss,rad[0]/(xyr*ss),cs, not fg,false,ci);

        { re-scale }
        if ss>1 then SImg_Upscale(cc, fk,fk, fi.nx,fi.ny);

        { subtraction }
        if high(sf)>=0 then
        begin
            r:=sf[0];
            for i:=0 to fi.size-1 do
                fk.pt[i]:=fi.pt[i]-fk.pt[i]*r;
        end;
    end
    else begin
        fl.pt:=nil; fm.pt:=nil;
        if irad>0 then SImg_Duplicate(cc,fk,fm,true)
        else SImg_Free(cc,fj);
        for j:=0 to high(rad) do
        begin
            { approximation }
            SImg_FBGFilter(cc, fk,fl, mask, rad[j]*xyr/ss,rad[j]/(xyr*ss),cs, not fg,false,ci);

            { subtraction }
            r:=sf[j];
            for i:=0 to fk.size-1 do
                fk.pt[i]-=fl.pt[i]*r;
        end;
        SImg_Free(cc, fl);
        if fm.pt<>nil then
        begin
            for i:=0 to fk.size-1 do
                fk.pt[i]:=fm.pt[i]-fk.pt[i];
            SImg_Free(cc, fm);
            if ss>1 then SImg_Upscale(cc, fk,fk, fi.nx,fi.ny);
            for i:=0 to fi.size-1 do
                fk.pt[i]:=fi.pt[i]-fk.pt[i];
        end;
    end;
    SImg_Move(cc,fk,fj);

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_NoiseFloor *************************************************}
{*************************************************************************}
{ Calculate the noise floor from a background corrected image
  fj : bg corrected: minimum filter with radius > 2*<superresolution factor> should be subtracted
}
function SImg_NoiseFloor_int(cc:CSImg_Config; fj: TSImg_FPImage):double;
var qs          : array[0..1] of double;
//    ps                : array[0..1] of double = (0.31, 0.69);         { without stars }
    ps          : array[0..1] of double = (0.30, 0.66);         { assuming with stars }
begin
    if not SImg_Recover(cc,fj) then cc.error('SImg_NoiseFloor_int: Invalid image: fj');
    SImg_GetQuantile(cc, fj, cc.threads,-1,ps,qs);
//  cc.info('SImg_NoiseFloor_int: '+real2str(qs[0])+'  '+real2str(qs[1]));
    result:=qs[1]-qs[0];
end;


{******* SImg_NoiseFloor *************************************************}
{ Calculate the noise floor
  fi    : input image
  sr    : superresolution factor >=1
}
function SImg_NoiseFloor(cc:CSImg_Config; fi:TSImg_FPImage; sr:double):double;
var fj          : TSImg_FPImage;
    i           : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_NoiseFloor: fi='+SImg_InfoStr2(fi)+' sr='+real2str(sr));
    if not SImg_Recover(cc,fi) then cc.error('SImg_NoiseFloor: Invalid image: fi');
    if sr<1 then sr:=1;
    sr:=sr+0.5;

    fj.pt:=nil;
    SImg_FBGFilter(cc, fi,fj, sr*2, true, false);
    for i:=0 to fi.size-1 do
        fj.pt[i]:=fi.pt[i]-fj.pt[i];
    result:=SImg_NoiseFloor_int(cc, fj);

    SImg_Free(cc,fj);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_FilterW ****************************************************}
{*************************************************************************}
threadprocedure SImg_FilterW_thread(ti:longint; tn:longint; fk:TSImg_FPImage; fj:TSImg_FPImage);
var i,ia,iz     : longint;
    a           : simgfp;
begin
    ia:=(fj.size*ti) div tn;
    iz:=(fj.size*(ti+1)) div tn-1;
    for i:=ia to iz do
    begin
        a:=fk.pt[i*2+1];
        if a<simg_mina then a:=simg_mina;
        fj.pt[i]:=fk.pt[i*2]/a;
    end;
end;

{ Gaussian filter with weight : exp(-sqr(r/rad)) }
procedure SImg_FilterW(cc:CSImg_Config; fi,fw:TSImg_FPImage; var fj:TSImg_FPImage; rad:double);
var fk  : TSImg_FPImage;
    i   : longint;
    r,s : double;
    a,b : simgfp;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_FilterW: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj)+' rad='+real2str(rad));
    if not SImg_Recover(cc,fi) then cc.error('SImg_FilterW: Invalid image: fi');
    if not SImg_Recover(cc,fw) then cc.error('SImg_FilterW: Invalid image: fw');
    if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>fi.nc) then cc.error('SImg_Filter: Invalid size: fw');
    if not SImg_OverwriteAlert(fi,fj) and not SImg_OverwriteAlert(fw,fj) then SImg_Free(cc,fj);

    fk.pt:=nil;
    SImg_New(cc, fk, fi.nx,fi.ny,fi.nc*2, false);

    { weight and regularization }
    r:=0; s:=0;
    for i:=0 to fw.size-1 do
    begin
        a:=fw.pt[i];
        if a<=0 then a:=0;
        fk.pt[i*2+1]:=a;
        r+=sqr(a);
        s+=a;
    end;
    if SImg_OverwriteAlert(fw,fj) and not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);
    if s<simg_mina then cc.error('SImg_FilterW: Weight image non-positive');
    a:=r/s*1e-9;
    if a<simg_mina then a:=simg_mina;

    { image*weight }
    for i:=0 to fi.size-1 do
    begin
        b:=fk.pt[i*2+1]+a;
        fk.pt[i*2+1]:=b;
        fk.pt[i*2  ]:=fi.pt[i]*b;
    end;
    if SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);

    { filter }
    SImg_Filter(cc, fk,fk, rad, 0, 1, 0.5);

    { output }
    SImg_New(cc, fj, fk.nx,fk.ny,fk.nc div 2, false);
    for i:=0 to cc.threads-1 do
        dothread SImg_FilterW_thread(i,cc.threads, fk, fj);
    threadwait;

    SImg_Free(cc, fk);
    SImg_DecDebugLevel;
end;
