{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{*************************************************************************}
{******* message stuff ***************************************************}
{*************************************************************************}
{*************************************************************************}
const SImg_DebugMessageWidth  = 320;
      SImg_MaxDebugLevel      = 128;

      SImg_DebugLevel         : longint = 0;
      SImg_ErrorMessageWidth  : longint = 320;

      SImg_DebugState         : longint = -1; { <0: not initialized; >0: running }

var SImg_DebugFileName  : simgstring;
    SImg_DebugFile      : longint;

//ifdef[timedebug]
type TSImg_DebugTime = record
        r,c     : double;       { real, CPU time }
        dr,dc   : double;       { real, CPU excluding sub routine measurements }
    end;
var SImg_DebugTimes     : array[0..SImg_MaxDebugLevel] of TSImg_DebugTime;
//endif


{*************************************************************************}
{******* fdwritestr ******************************************************}
{*************************************************************************}
function fdwritestr(fd:longint; const str:ansistring):boolean;  { true on success }
begin
    result:=fpwrite(fd,str[1],length(str))=length(str);
end;


{*************************************************************************}
{******* SImg_Info *******************************************************}
{*************************************************************************}
procedure SImg_Info(const s:ansistring);
begin
    if SImg_Config_Default=nil
    then SImg_StdPrintMessage(-3,s,true)
    else SImg_Config_Default.Info(s);
end;


{*************************************************************************}
{******* SImg_Hint *******************************************************}
{*************************************************************************}
procedure SImg_Hint(const s:ansistring);
begin
    if SImg_Config_Default=nil
    then SImg_StdPrintMessage(-2,s,true)
    else SImg_Config_Default.Hint(s);
end;


{*************************************************************************}
{******* SImg_Warning ****************************************************}
{*************************************************************************}
procedure SImg_Warning(const s:ansistring);
begin
    if SImg_Config_Default=nil
    then SImg_StdPrintMessage(-1,s,true)
    else SImg_Config_Default.Warning(s);
end;


{*************************************************************************}
{******* SImg_Error ******************************************************}
{*************************************************************************}
procedure SImg_Error(const s:ansistring);
begin
    if SImg_Config_Default=nil then
    begin
        SImg_StdPrintMessage(0,s,true);
        raise CSImg_ErrorException.create(s);
    end
    else SImg_Config_Default.Error(s);
end;


{*************************************************************************}
{******* SImg_FatError ***************************************************}
{*************************************************************************}
procedure SImg_FatError(const s:ansistring);
begin
    if SImg_Config_Default=nil then
    begin
        SImg_StdPrintMessage(1,s,true);
        SImg_Halt(2);
    end
    else SImg_Config_Default.FatError(s);
end;


{*************************************************************************}
{******* SImg_MatrixError ************************************************}
{*************************************************************************}
procedure SImg_MatrixError(i:longint; const s:string);
begin
    case i of
        0 : SImg_Error(s);
        1 : SImg_Warning(s);
        else SImg_Hint(s);
    end;
end;


{*************************************************************************}
{******* SImg_FormatDebugString ******************************************}
{*************************************************************************}
{
  mes  : message string
  len  : maximal length of line (infinity if len<1)
  b1   : spaces after line break
  b2   : extra spaces after automatic line break (see len)
}
function SImg_FormatDebugString(const mes:ansistring; len,b1,b2:longint):ansistring;
var i,j,k,l,m  : longint;
    s,sb1,sb2  : ansistring;
begin
    sb1:='';
    for i:=1 to b1 do
        sb1+=' ';
    sb2:='';
    for i:=1 to b1+b2 do
        sb2+=' ';
    s:='';
    j:=0; m:=1;
    for i:=1 to length(mes) do
    begin
        case mes[i] of
            #9 :
                repeat
                    s+=' ';
                    j+=1;
                until j and 7=0;
            #13 :
                begin
                    s+=#13;
                    j:=0;
                    m:=length(s)+1;
                end;
            #10 :
                begin
                    s+=#10+sb1;
                    j:=b1;
                    m:=length(s)+1;
                end;
            else
                begin
                    s+=mes[i];
                    j+=1;
                end;
         end;
        if (len>0) and (j>len) and (s[length(s)]<>#32)  then
        begin
            k:=length(s);
            while (s[k]<>#32) and (k>m) do
                k-=1;
            if s[k]=#32 then
            begin
                l:=k;
                while (l>m) and (s[l-1]=#32) do
                    l-=1;
                delete(s,l,k-l+1);
                insert(lineEnding+sb2,s,l);
                j:=length(s)-l;
                m:=l+b1+b2+1;
            end;
        end;
    end;
    k:=length(s);
    while (k>1) and (s[k]=#32) do
        k-=1;
    if k<length(s) then delete(s,k+1,length(s)-k);
    SImg_FormatDebugString:=s;
end;


{*************************************************************************}
{******* SImg_DebugMessage ***********************************************}
{*************************************************************************}
var debugspaces : array[0..SImg_MaxDebugLevel] of longint;

procedure SImg_DebugMessage(const s:ansistring);
var u       : shortstring;
begin
    if SImg_DebugState<=0 then exit;
    if SImg_DebugLevel<0 then
    begin
        SImg_DebugLevel:=-1;
        SImg_IncDebugLevel;
    end;
    setlength(u,SImg_DebugLevel*2);
    fillchar(u[1],length(u),' ');
    u:=formatDateTime('HH:NN:SS.zzz',now)+{' '+int2str(gettid)+}' ('+int2str(SImg_DebugLevel)+'): '+u;
    debugspaces[SImg_DebugLevel]:=length(u)+4;
{$I+}
    try
        fdwritestr(SImg_DebugFile,SImg_FormatDebugString(u+s,SImg_DebugMessageWidth,debugspaces[SImg_DebugLevel],0)+lineEnding);
    except
        on E: EInOutError do
            fdwritestr(2,'Warning: SImg_DebugMessage: Unable to write to debug file: '+E.Message);
      end;
{$I-}
end;

(*procedure SImg_DebugMessage2(const s:ansistring);
var i   : longint;
    t   : shortstring;
begin
    if SImg_DebugState<=0 then exit
    i:=debugspaces[SImg_DebugLevel];
    setlength(t,i);
    fillchar(t,length(t),' ');
{$I+}
    try
        fdwritestr(SImg_DebugFile,SImg_FormatDebugString(t+s,SImg_DebugMessageWidth,i,0)+lineEnding)
    except
        on E: EInOutError do
            fdwritestr(2,'Warning: SImg_DebugMessage: Unable to write to debug file: '+E.Message);
    end;
{$I-}
end; *)


{*************************************************************************}
{******* SImg_IncDebugLevel **********************************************}
{*************************************************************************}
procedure SImg_IncDebugLevel;
//ifdef[timedebug]
var tv  : timespec;
//endif
begin
    if SImg_DebugState<=0 then exit;
    if SImg_DebugLevel<SImg_MaxDebugLevel then
    begin
        inc(SImg_DebugLevel);
//ifdef[timedebug]
        with SImg_DebugTimes[SImg_DebugLevel] do
        begin
            clock_gettime(CLOCK_MONOTONIC_RAW,@tv);
            r:=tv.tv_sec + tv.tv_nsec*1e-9;
            dr:=r;
            clock_gettime(CLOCK_PROCESS_CPUTIME_ID,@tv);
            c:=tv.tv_sec + tv.tv_nsec*1e-9;
            dc:=c;
        end;
//endif
    end;
end;

{*************************************************************************}
{******* SImg_DecDebugLevel **********************************************}
{*************************************************************************}
procedure SImg_DecDebugLevel;
//ifdef[timedebug]
var d   : double;
    s   : ansistring;
    i   : longint;
    tv  : timespec;
//endif
begin
    if (SImg_DebugState<=0) or (SImg_DebugLevel<0) then exit;
//ifdef[timedebug]
    with SImg_DebugTimes[SImg_DebugLevel] do
    begin
        clock_gettime(CLOCK_MONOTONIC_RAW,@tv);
        d:=tv.tv_sec + tv.tv_nsec*1e-9;
        r:=d-r;
        dr:=d-dr;
        clock_gettime(CLOCK_PROCESS_CPUTIME_ID,@tv);
        d:=tv.tv_sec + tv.tv_nsec*1e-9;
        c:=d-c;
        dc:=d-dc;
        s:='Total timing: real='+time2str(r) + '  cpu='+time2str(c);
        if (dc<=0.9999*c) or (dr<0.9999*r) then s+='      Timing w/o sub measurements:  real='+time2str(dr) + '  cpu='+time2str(dc);
        SImg_debugMessage(s);
    end;
    for i:=0 to SImg_DebugLevel-1 do
    begin
        SImg_DebugTimes[i].dr+=SImg_DebugTimes[SImg_DebugLevel].dr;
        SImg_DebugTimes[i].dc+=SImg_DebugTimes[SImg_DebugLevel].dc;
    end;
//endif
    dec(SImg_DebugLevel);
end;


{*************************************************************************}
{******* SImg_DebugInit **************************************************}
{*************************************************************************}
procedure SImg_GetDebugFileName(cc:CSImg_Config; i:longint; var s:shortstring);
begin
    s:=int2str(i);
    if (cc.logFileNum>9) and (length(s)<2) then s:='0'+s;
    s:=SImg_HomeDir+'/'+cc.LogFileBase+'.'+s+'.log';
end;

procedure SImg_DebugInit;
var dir            : pdir;
    s,s2           : shortstring;
    i,j            : longint;

begin
    if (SImg_DebugState>=0) or (SImg_Config_Default=nil) or (SImg_Config_Default.logFileNum<0) then exit;
    SImg_DebugState:=0;

    SImg_HomeDir:=getEnvironmentVariable('HOME');
    SImg_HomeDir+='/'+SImg_Dir;
    dir:=fpOpenDir(SImg_HomeDir);
    if dir=nil then
    begin
        if not CreateDir(SImg_HomeDir) then
        begin
            fdwritestr(2,'SImg_InitSys: Can''t create directory: '+SImg_HomeDir);
            exit;
         end;
    end
    else fpCloseDir(dir^);

    i:=-1;
    s:='';
    repeat
        i+=1;
        SImg_GetDebugFileName(SImg_Config_Default,i,s);
    until (i>=SImg_Config_Default.logFileNum) or (not FileExists(s));
    if i>=SImg_Config_Default.LogFileNum then
    begin
        i-=1;
        SImg_GetDebugFileName(SImg_Config_Default,i,s);
    end;
    for j:=i-1 downto 0 do
    begin
        SImg_GetDebugFileName(SImg_Config_Default,j,s2);
        RenameFile(s2,s);
        s:=s2;
    end;

    SImg_DebugFile:=fpopen(s,Open_Creat or Open_Trunc or Open_WROnly);
    if (SImg_DebugFile<0) then
    begin
        fdwritestr(2,'SImg_InitSys: Unable to create debug file: '+s);
        SImg_DebugFileName:='';
        exit;
    end;

    SImg_DebugState:=1;
    SImg_DebugMessage('SImg_InitSys: started: '+formatDateTime('YYYY-MM-DD',now)+', TID='+int2str(gettid)+', using up to '+int2str(SImg_Config_Default.threads)+' threads per default');

    SImg_DebugLevel:=-1;
    SImg_IncDebugLevel;
end;


{*************************************************************************}
{******* SImg_StdPrintMessage ********************************************}
{*************************************************************************}
{ t : type of message:
       <-3 : debug
       -3  : Info
       -2  : Hint
       -1  : Warning
        0  : Error
       >0  : Fatal
  mes  : message String
  print: write to stdout }
procedure SImg_StdPrintMessage(t:longint; const mes:ansistring; print:boolean);
begin
    if print then
    begin
        if t=-1 then fdwritestr(2,SImg_FormatDebugString('Warning: '+mes,SImg_ErrorMessageWidth,9,0)+lineEnding)
        else if t=0 then fdwritestr(2,SImg_FormatDebugString('Error: '+mes,SImg_ErrorMessageWidth,7,0)+lineEnding)
        else if t>0 then fdwritestr(2,SImg_FormatDebugString('Fatal Error: '+mes,SImg_ErrorMessageWidth,7,0)+lineEnding)
        else fdwritestr(2,SImg_FormatDebugString(mes,SImg_ErrorMessageWidth,0,0)+lineEnding);
    end;

    if SImg_DebugState<0 then SImg_DebugInit;
    if SImg_DebugState<=0 then exit;

    if t<-3 then SImg_DebugMessage(mes)
    else if t=-3 then SImg_DebugMessage('Info: ' + mes)
    else if t=-2 then SImg_DebugMessage('Hint: ' + mes)
    else if t=-1 then SImg_DebugMessage('Warning: ' + mes)
    else if t=0 then SImg_DebugMessage('Error: ' + mes)
    else SImg_DebugMessage('Fatal Error: ' + mes)
end;


{*************************************************************************}
{******* SImg_InitMessages ***********************************************}
{*************************************************************************}
procedure SImg_InitMessages;
var ws             : twinsize;
begin
    if (isatty(stderr)>0) and (fpioctl(textrec(stderr).handle,TIOCGWINSZ,@ws)>=0) then SImg_ErrorMessageWidth:=ws.ws_col;
    if SImg_ErrorMessageWidth<80 then SImg_ErrorMessageWidth:=80;
end;


{*************************************************************************}
{******* SImg_DoneMessages ***********************************************}
{*************************************************************************}
procedure SImg_DoneMessages(ec:longint);
begin
    if SImg_DebugState<=0 then exit;
    if ec=0 then
    begin
        while SImg_debugLevel>=0 do
            SImg_DecDebugLevel;
    end;
    SImg_DebugMessage('SImg_DoneSys: called with: exitcode='+int2str(ec));
    fpclose(SImg_DebugFile);
end;
