{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{*************************************************************************}
{******* important system routines ***************************************}
{*************************************************************************}
{*************************************************************************}
type TSImg_MemoryBlock=record
      case smallint of
        0 : ( first,last  : qword;
//ifdef[memdebug]
              caller_addr : pointer;
//endif
            );
        1 : ( pt          : pointer );
      end;

const SImg_MemoryBlocks_InitialSize = 8192;
      SImg_MemoryBlocks_Size : longint = SImg_MemoryBlocks_InitialSize;
      SImg_MemoryBlocks  : ^TSImg_MemoryBlock = nil;

//ifdef[memdebug]
const SImg_AllocatedMemory    : int64 = 0;
      SImg_MaxAllocatedMemory : aword = 0;
//endif

var disable_unhandledException : boolean;

{*************************************************************************}
{******* SImg_GetLineInfo ************************************************}
{*************************************************************************}
function SImg_GetLineInfo(addr:pointer):ansistring;
begin
    if SImg_Config_Default<>nil then SImg_Config_Default.debug('SImg_GetLineInfo: addr='+hexstr(longword(addr),sizeof(addr)*2));
    result:=SysBackTraceStr(addr);
end;

{*************************************************************************}
{******* SImg_GetStackTrace **********************************************}
{*************************************************************************}
const SImg_MaxStackDumps = 7;
function SImg_GetStackTrace(frame:pointer):ansistring;
var i   : longint;
    s   : ansistring;
    a   : pointer;
begin
    i:=0;
    s:='';
    while (i<SImg_MaxStackDumps) and (frame<>nil) do
    begin
        a:=get_caller_addr(frame);
        if SImg_Config_Default<>nil then SImg_Config_Default.debug('SImg_GetLineInfo: addr='+hexstr(longword(a),sizeof(a)*2));
        if i>0 then s+=lineEnding+'    ';
        s+=SysBackTraceStr(a);
        frame:=get_caller_frame(frame);
        i+=1;
    end;
    result:=s;
end;

{*************************************************************************}
{******* SImg_ExceptionToMessage *****************************************}
{*************************************************************************}
function SImg_ExceptionToMessage(obj:tobject; addr:pointer; framecount:longint; frames:ppointer):ansistring;
var s         : simgstring;
    i,j       : longint;
begin
    if obj is exception then s:=exception(obj).message
    else s:='Invalid exception object: '+obj.classname;
//ifdef[debug]
    if addr<>nil then s+=lineEnding+'    '+SImg_GetLineInfo(addr);
    j:=framecount;
    if SImg_MaxStackDumps<j then j:=SImg_MaxStackDumps;
    for i:=0 to j-1 do
        if frames[i]<>nil then s+=lineEnding+'    '+SImg_GetLineInfo(frames[i]);
//endif
    SImg_ExceptionToMessage:=s;
end;

function SImg_ExceptionToMessage:simgstring;
var  eo  : PExceptObject;
begin
    eo:=raiselist;
    if eo<>nil then SImg_ExceptionToMessage:=SImg_ExceptionToMessage(eo^.fobject, eo^.addr, eo^.framecount, eo^.frames)
    else SImg_ExceptionToMessage:='';
end;


{*************************************************************************}
{******* SImg_QuiteException *********************************************}
{*************************************************************************}
procedure SImg_QuiteException;
begin
    raise CSImg_QuiteException.create('');
end;


{*************************************************************************}
{******* SImg_UnhandledException *****************************************}
{*************************************************************************}
procedure SImg_UnhandledException(obj:tobject; addr:pointer; framecount:longint; frames:ppointer);
var s   : simgstring;
    b   : boolean;
    mp  : TSImg_PrintMessageProc;
begin
    if disable_unhandledException then exit;
    disable_unhandledException:=true;
    b:=SImg_Sem_reset;
    if b then write(stderr,'Trying to find out more info ...');
    if (SImg_Config_Default<>nil) and SImg_Config_Default.MessageProcSet then mp:=SImg_Config_Default.MessageProc
    else mp:=@SImg_StdPrintMessage;

    if obj is CSImg_ErrorException then mp(1,'Unhandled Exception occurred: '+CSImg_ErrorException(obj).message,true)
    else if obj is CSImg_QuiteException then mp(-4,'SImg_UnhandledExeption: CSImg_QuiteException',true)
    else begin
       s:=SImg_ExceptionToMessage(obj,addr,framecount,frames);
       mp(1,'Unhandled Exception occurred: '+s,true);
    end;
if b then SImg_Halt(2)
     else SImg_Halt(1);
end;

{*************************************************************************}
{******* SImg_OutputCatchedException *************************************}
{*************************************************************************}
procedure SImg_OutputCatchedException(cc:CSImg_Config);
var eo  : PExceptObject;
    obj : tobject;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_Sem_Reset;
    eo:=raiselist;
    if eo=nil then exit;
    obj:=eo^.fobject;
    if (obj=nil) or (obj is CSImg_ErrorException) or (obj is CSImg_QuiteException) then exit;

    cc.debug('SImg_OutputCatchedException: '+SImg_ExceptionToMessage);
    if obj is exception
    then cc.PrintMessage(0,'Exception: '+exception(obj).message,true)
    else cc.PrintMessage(0,'Exception: Invalid exception object: '+obj.classname,true);
end;


{*************************************************************************}
{******* SImg_CheckMemoryIntegrity ***************************************}
{*************************************************************************}
//ifdef[memdebug]
function SImg_CheckMemoryIntegrity:boolean;
var i,j  : longint;
    k    : int64;
begin
    j:=0; k:=0;
    for i:=0 to SImg_MemoryBlocks_Size-1 do
        with SImg_MemoryBlocks[i] do
            if first=0 then
            begin
                if last<>0 then inc(j);
            end
            else begin
                if last<first then inc(j);
//              if last<first then Simg_Warning(base1024str(first)+' >  '+base1024str(last));
                k+=last-first+1;
            end;
    if k<>SImg_AllocatedMemory then inc(j);
//  if k<>SImg_AllocatedMemory then Simg_Warning(base1024str(k)+' <>  '+base1024str(SImg_AllocatedMemory));
    if j>0 then SImg_Warning('SImg_CheckMemoryIntegrity: Inconsistency in memory allocation table: detected '+int2str(j)+' errors');
    SImg_CheckMemoryIntegrity:=j=0;
end;
//endif

{*************************************************************************}
{******* getmem2  ********************************************************}
{*************************************************************************}
//ifdef[cmem]
function malloc(size:Longint):Pointer; cdecl; external 'c' name 'malloc';
procedure free(p:pointer); cdecl; external 'c' name 'free';
//endif

procedure getmem2(var ptr:pointer;size:aword);
var cc   : CSIMg_Config;
    i,j  : longint;
    a,b  : aword;
    tmp  : ^TSImg_MemoryBlock;

begin
    cc:=SImg_Config_Default;
//ifdef[xmemdebug]
    cc.debug('getmem2(1): called with: ptr='+ptr2str(ptr)+' size='+base1024str(size));
//endif
    if size<=0 then cc.warning('getmem2: Can''t allocate 0 bytes of memory');
    SImg_Sem_Lock(SImg_Sem_Mem);
    try
        try
//ifdef[cmem]
            ptr:=malloc(size);
//else
            getmem(ptr,size);
//endif
         except
            ptr:=nil;
        end;
//ifdef[memdebug]
        if ptr=nil then cc.Error('getmem2: Can''t allocate '+base1024str(size)+' bytes of memory, '+base1024str(SImg_AllocatedMemory)+' are allocated now');
//else
        if ptr=nil then cc.Error('getmem2: Can''t allocate '+base1024str(size)+' bytes of memory');
//endif

        i:=0;
        while (SImg_MemoryBlocks[i].first<>0) and (i<SImg_MemoryBlocks_size) do
            inc(i);
        if i=SImg_MemoryBlocks_Size then
        begin
            tmp:=SImg_MemoryBlocks;
            SImg_MemoryBlocks_Size+=SImg_MemoryBlocks_InitialSize;
            getmem(SImg_MemoryBlocks,SImg_MemoryBlocks_Size*sizeof(TSImg_MemoryBlock));
            move(tmp^,SImg_MemoryBlocks^,i*sizeof(TSImg_MemoryBlock));
            freemem(tmp,i*sizeof(TSImg_MemoryBlock));
            for j:=i to SImg_MemoryBlocks_Size-1 do
            with SImg_MemoryBlocks[j] do
            begin
                first:=0;
                last:=0;
            end;
        end;
        with SImg_MemoryBlocks[i] do
        begin
            pt:=ptr;
            last:=first+size-1;
            a:=first;
            b:=last;
//ifdef[memdebug]
            caller_addr:=get_caller_addr(get_frame);
//endif
        end;
//ifdef[memdebug]
        for j:=0 to SImg_MemoryBlocks_Size-1 do
            if (i<>j) and (SImg_MemoryBlocks[j].first<>0) then
                with SImg_MemoryBlocks[j] do
                    if (b>=first) and (a<=last) then cc.Warning('getmem2: Memory block overlap detected');
        SImg_AllocatedMemory+=size;
        if SImg_AllocatedMemory>SImg_MaxAllocatedMemory then SImg_MaxAllocatedMemory:=SImg_AllocatedMemory;
        SImg_CheckMemoryIntegrity;
//endif
    finally
        SImg_Sem_Unlock(SImg_Sem_Mem);
    end;
//ifdef[memdebug]
//ifdef[xmemdebug]
    cc.debug('getmem2(2): ','allocated '+base1024str(size)+' bytes at '+ptr2str(ptr)+ ', '+base1024str(SImg_AllocatedMemory)+' bytes are allocated now');
//endif
//endif
end;

{*************************************************************************}
{******* freemem2  *******************************************************}
{*************************************************************************}
procedure freemem2(var ptr:pointer;size:aword);
var cc   : CSIMg_Config;
    opt : pointer;
    a    : aword absolute opt;
    b    : aword;
    i    : longint;
begin
    cc:=SImg_Config_Default;
//ifdef[xmemdebug]
    cc.debubug('feeemem2(1): called with: ptr='+ptr2str(ptr)+' size='+base1024str(size));
//endif
    SImg_Sem_Lock(SImg_Sem_Mem);
    try
        opt:=ptr;
        b:=a+size-1;
        i:=0;
        while ((SImg_MemoryBlocks[i].pt<>ptr) or (SImg_MemoryBlocks[i].first=0)) and (i<SImg_MemoryBlocks_Size) do
            inc(i);
        if i>=SImg_MemoryBlocks_Size then
        begin
//ifdef[memdebug]
            i:=0;
            while ((b<SImg_MemoryBlocks[i].first) or (a>SImg_MemoryBlocks[i].last)) and (i<SImg_MemoryBlocks_Size) do
                inc(i);
            if i>=SImg_MemoryBlocks_Size
            then cc.warning('freemem2: Deallocating unknown memory block')
            else cc.warning('freemem2: Deallocating illegal memory block');
//endif
        end
        else with SImg_MemoryBlocks[i] do
        begin
//ifdef[memdebug]
            if b<>last then
            begin
                cc.warning('freemem2: Deallocating invalid memory block');
                size:=last-first+1;
            end;
//endif
        first:=0; last:=0;
        end;
//ifdef[cmem]
        free(ptr);
//else
        freemem(ptr,size);
//endif
//ifdef[memdebug]
        SImg_AllocatedMemory-=size;
        SImg_CheckMemoryIntegrity;
//endif
    finally
        SImg_Sem_Unlock(SImg_Sem_Mem);
        ptr:=nil;
    end;

//ifdef[memdebug]
//ifdef[xmemdebug]
    cc.debug('freemem2(2): deallocated '+base1024str(size)+' bytes at '+ptr2str(opt)+', '+base1024str(SImg_AllocatedMemory)+' bytes are allocated now');
//endif
    if SImg_AllocatedMemory<0 then cc.warning('freemem2: Invalid memory balance');
//endif
end;

procedure freemem2(var ptr:pointer);
var cc   : CSIMg_Config;
    b    : aword;
    i    : longint;
//ifdef[xmemdebug]
    opt  : pointer;
//endif
begin
    cc:=SImg_Config_Default;
//ifdef[xmemdebug]
    cc.debug('feeemem2(1): ','called with: ptr='+ptr2str(ptr));
//endif
    SImg_Sem_Lock(SImg_Sem_Mem);
    try
//ifdef[xmemdebug]
        opt:=ptr;
//endif
        i:=0;
        while ((SImg_MemoryBlocks[i].pt<>ptr) or (SImg_MemoryBlocks[i].first=0)) and (i<SImg_MemoryBlocks_Size) do
            inc(i);
        if i>=SImg_MemoryBlocks_Size
        then cc.warning('freemem2: Impossible to deallocate unknown memory block')
        else with SImg_MemoryBlocks[i] do
        begin
            b:=last-first+1;
//ifdef[cmem]
            free(ptr);
//else
            freemem(ptr,b);
//endif
            ptr:=nil;
            first:=0; last:=0;
//ifdef[memdebug]
            SImg_AllocatedMemory-=b;
            SImg_CheckMemoryIntegrity;
//ifdef[xmemdebug]
            cc.debug('freemem2(2): ','deallocated '+base1024str(b)+' bytes at '+ptr2str(opt)+', '+base1024str(SImg_AllocatedMemory)+' bytes are allocated now');
//endif
            if SImg_AllocatedMemory<0 then SImg_Warning('freemem2: Invalid memory balance');
//endif
        end;
    finally
        SImg_Sem_Unlock(SImg_Sem_Mem);
    end;
end;


{*************************************************************************}
{******* SImg_InitSys ****************************************************}
{*************************************************************************}
procedure SImg_InitSys;
var i  : longint;
begin
    getmem(SImg_MemoryBlocks,SImg_MemoryBlocks_Size*sizeof(TSImg_MemoryBlock));
    for i:=0 to SImg_MemoryBlocks_size-1 do
    with SImg_MemoryBlocks[i] do
    begin
        first:=0;
        last:=0;
    end;
    ExceptProc:=@SImg_UnhandledException;
    InitErrorProc(@SImg_MatrixError);
    disable_unhandledException:=false;
end;


{*************************************************************************}
{******* SImg_DoneSys ****************************************************}
{*************************************************************************}
procedure SImg_DoneSys(ec:longint);
var cc   : CSIMg_Config;
    i    : longint;
begin
    cc:=SImg_Config_Default;
    if SImg_DebugState<0 then SImg_DebugState:=0;
//ifdef[memdebug]
    if (ec=0) and (SImg_AllocatedMemory<>0) then
    begin
        SImg_warning('SImg_DoneSys: Still memory allocated');
        for i:=0 to SImg_MemoryBlocks_Size-1 do
            with SImg_MemoryBlocks[i] do
                if first<>0 then SImg_Warning('SImg_DoneSys: Allocated memory ('+base1024str(last-first+1)+' bytes) at '+ptr2str(pt)+', allocated by '+SImg_GetLineInfo(caller_addr));
    end;
//endif
    RestoreErrorProc;

//ifdef[memdebug]
    cc.debug('SImg_DoneSys: '+base1024str(SImg_AllocatedMemory)+' bytes are allocated');
    cc.debug('SImg_DoneSys: '+base1024str(SImg_MaxAllocatedMemory)+' bytes where allocated at maximum');
//endif
//disbale_unhandledException:=true;
    freemem(SImg_MemoryBlocks,SImg_MemoryBlocks_Size*sizeof(TSImg_MemoryBlock));
end;

