{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{*************************************************************************}
{******* useful routines *************************************************}
{*************************************************************************}
{*************************************************************************}

{*************************************************************************}
{******* conversion of real to string ************************************}
{*************************************************************************}
{******* real2str_fixed_length *******************************************}
{******* real2str ********************************************************}
{ dig: number of valid digits
    <=0   : -number of digits behind comma
    1..15 : absolute number of digits
    >5    : raw (freepascal) output }
function real2str(d:extended; dig:longint):string31;
var i,j    : longint;
    k,l    : int64;
    s      : shortstring;
begin
    if dig=0 then
    begin
        result:=int2str(round(d));
        exit;
    end;
    if dig<0 then
    begin
        str(d:0:-dig,result);
        exit;
    end;
    if dig>15 then
    begin
        str(d,result);
        exit;
    end;
    if abs(d)<1e-4900 then
    begin
        result:='0';
        if dig>1 then result+='.';
        for i:=2 to dig do
            result+='0';
        exit;
    end;
    if d<0 then
    begin
        result:='-';
        d:=-d;
    end
    else result:='';

    l:=1;
    for i:=2 to dig do
        l*=10;

    d*=l;
    i:=0;
    while d<l do
    begin
        d*=10;
        i-=1;
    end;

    l*=10;
    while d>=l do
    begin
        d*=0.1;
        i+=1;
    end;
    k:=round(d);
    if k>=l then
    begin
        k:=(k+5) div 10;
        i+=1;
    end;

    s:='';
    for j:=1 to dig do
    begin
        s:=char(k mod 10+48)+s;
        k:=k div 10;
    end;
    if i>=100 then
    begin
        if dig>1 then insert('.',s,2);
        s+='E+'+char(i div 100+48)+char((i div 10) mod 10+48)+char(i mod 10+48);
    end
    else if i>=10 then
    begin
        if dig>1 then insert('.',s,2);
        s+='E+'+char(i div 10+48)+char(i mod 10+48);
    end
    else if i>=dig then
    begin
        if dig>1 then insert('.',s,2);
        s+='E+'+char(i+48);
    end
    else if i=dig-1 then
    begin
    end
    else if i>=0 then insert('.',s,i+2)
    else if i=-1 then insert('0.',s,1)
    else if i=-2 then insert('0.0',s,1)
    else if i>-10 then
    begin
        if dig>1 then insert('.',s,2);
        i:=-i;
        s+='E-'+char(i+48);
    end
    else if i>-100 then
    begin
       if dig>1 then insert('.',s,2);
       i:=-i;
       s+='E-'+char(i div 10+48)+char(i mod 10+48);
    end else
    begin
        if dig>1 then insert('.',s,2);
        i:=-i;
        s+='E-'+char(i div 100+48)+char((i div 10) mod 10+48)+char(i mod 10+48);
    end;
    result+=s;
end;

function real2str(d:extended):string31;
begin
    result:=real2str(d,4);
end;

{******* time2str ********************************************************}
function time2str(t:extended):string15;
var u  : string[3];
    s  : boolean;
    i  : longint;
begin
    s:=t<0;
    t:=abs(t);
    if t>600 then
    begin
        t/=60;
        u:='m';
    end
    else if t>1 then u:='s'
    else if t>1e-3 then
    begin
        t*=1000;
        u:='ms';
    end
    else if t>1e-6 then
    begin
        t*=1e6;
        u:='µs';
    end else
    begin
        t*=1e9;
        u:='ns';
    end;
    if t<10 then i:=4
    else if t<100 then i:=3
    else i:=2;

    str(t:0:i,result);
    if s then result:='-'+result+u
    else result+=u;
end;



{*************************************************************************}
{******* conversion of integer to string *********************************}
{*************************************************************************}
{******* int2str *********************************************************}
{ convert integer to string }
function int2str(i:int64):string15;
begin
    str(i,int2str);
end;

function int2str(i:int64;j:longint):string15;
begin
    str(i:j,int2str);
end;

{******* base1024str *****************************************************}
{ convert integer to string in that way: 15385 -> 5K265 }
function base1024str(d:qword):string31;
var s,t : string;
begin
    s:='';
    if d and 1023 > 0 then
    begin
        str(d and 1023,t);
        s:=t;
    end;
    d:=d shr 10;
    if d and 1023 > 0 then
    begin
       str(d and 1023,t);
       s:=t+'K'+s;
    end;
    d:=d shr 10;
    if d and 1023 > 0 then
    begin
       str(d and 1023,t);
       s:=t+'M'+s;
    end;
    d:=d shr 10;
    if d and 1023 > 0 then
    begin
       str(d and 1023,t);
       s:=t+'G'+s;
    end;
    d:=d shr 10;
    if d and 1023 > 0 then
    begin
        str(d and 1023,t);
        s:=t+'T'+s;
    end;
    if s='' then s:='0';
    base1024str:=s;
end;

{******* hex2str *********************************************************}
{ convert integer to string (hex output) }
const hexchars:string[16]='0123456789ABCDEF';

function hex2str(i:dword):string15;
begin
    hex2str:='';
    repeat
        hex2str:=hexchars[i and 15]+hex2str;
        i:=i shr 4;
    until i=0;
end;


{*************************************************************************}
{******* other conversions ***********************************************}
{*************************************************************************}
{******* ptr2str *********************************************************}
{ convert pointer to string (hex output) }
function ptr2str(pt:pointer):string15;
{$ifdef CPU64}
var d:qword absolute pt;
    i:longint;
begin
    ptr2str[0]:=#15;
    ptr2str[1]:='$';
    for i:=0 to 13 do
    begin
        ptr2str[15-i]:=hexchars[(d and 15)+1];
{$else}
var d:dword absolute pt;
    i:longint;
begin
    ptr2str[0]:=#9;
    ptr2str[1]:='$';
    for i:=0 to 7 do
    begin
        ptr2str[9-i]:=hexchars[(d and 15)+1];
{$endif}
        d:=d shr 4;
 end;
end;

{******* booelan2str *****************************************************}
{ convert boolean to string (True/False) }
function boolean2str(b:boolean):string7;
begin
    if b then boolean2str:='True'
    else boolean2str:='False';
end;

{******* onoff2str *******************************************************}
{ convert boolean to string (on/off) }
function onoff2str(b:boolean):string7;
begin
    if b then onoff2str:='on'
    else onoff2str:='off';
end;


{*************************************************************************}
{******* misc ************************************************************}
{*************************************************************************}
{******* upstr ***********************************************************}
{ convert string to uppercase letters }
procedure upstr(var s:shortstring);
var i:longint;
begin
    for i:=1 to length(s) do
        s[i]:=upcase(s[i]);
end;

{******* min/max *********************************************************}
{ minimum/ maximum of two integers }
function min(a,b:longint):longint;inline;
begin
    if a<b then min:=a
    else min:=b;
end;

function max(a,b:longint):longint;inline;
begin
    if a>b then max:=a
    else max:=b;
end;

{******* condstring ******************************************************}
{ conditional string:
   b=true  : result:=s1
   b=false : result:=s2 }
function condstring(b:boolean;s1,s2:shortstring):shortstring;
begin
    if b then result:=s1
    else result:=s2;
end;


{******* trimstr *********************************************************}
{ clear spaces
   l:leading
   e:ending }
procedure trimstr(var s:ansistring; l,e:boolean);
var i : longint;
begin
    if l then
    begin
        i:=1;
        while (i<=length(s)) and (s[i]<=#32) do
            inc(i);
        if i>1 then delete(s,1,i-1);
    end;
    if e then
    begin
        i:=length(s);
        while (i>0) and (s[i]<=#32) do
            dec(i);
        if i<length(s) then delete(s,i+1,length(s)-i);
    end;
end;

procedure trimstr(var s:shortstring; l,e:boolean);
var i : longint;
begin
    if l then
    begin
        i:=1;
        while (i<=length(s)) and (s[i]<=#32) do
            inc(i);
        if i>1 then delete(s,1,i-1);
    end;
    if e then
    begin
        i:=length(s);
        while (i>0) and (s[i]<=#32) do
            dec(i);
        if i<length(s) then delete(s,i+1,length(s)-i);
    end;
end;

procedure trimstr(var s:ansistring);
begin
    trimstr(s,true,true);
end;

procedure trimstr(var s:shortstring);
begin
    trimstr(s,true,true);
end;


{******* trimquotes ********************************************************}
function trimquotes(var s:ansistring):ansistring;
var i,j : longint;
begin
    i:=1;
    while (i<=length(s)) and (s[i]<=#32) do
        inc(i);
    if (i<=length(s)) and (s[i]='''') then inc(i);

    j:=length(s);
    while (j>=i) and (s[j]<=#32) do
        dec(j);
    if (j>=i) and (s[j]='''') then dec(j);
    if j>=i then result:=copy(s,i,j-i+1)
    else result:='';
end;

{******* min ***************************************************************}
function min(const a,b:double):double; inline;
begin
    if a<b then result:=a
    else result:=b;
end;

{******* max ***************************************************************}
function max(const a,b:double):double; inline;
begin
    if a>b then result:=a
    else result:=b;
end;

{******* str2pchar *********************************************************}
function str2pchar(var str:shortstring):pchar;
var i  : longint;
begin
    i:=length(str)+1;
    if i>255 then i:=255;
    str[i]:=#0;
    result:=@str[1];
end;

function str2pchar(var str:ansistring):pchar;
begin
    result:=@str[1];
end;

{******* gettid ***********************************************************}
function gettid:longint;
begin
    result:=do_syscall(syscall_nr_gettid);
end;

{******* val2 *************************************************************}
{ return false on error }
function val2(s:shortstring; var d:double):boolean;
var w   : word;
    a,b : double;
begin
    a:=1.0;
    if s[length(s)]='%' then
    begin
        a:=0.01;
        delete(s,length(s),1);
    end;
    val(s,b,w);
    result:=w<=0;
    if result then d:=a*b;
end;

function val2(s:shortstring; var d:single):boolean;
var w   : word;
    a,b : double;
begin
    a:=1.0;
    if s[length(s)]='%' then
    begin
        a:=0.01;
        delete(s,length(s),1);
    end;
    val(s,b,w);
    result:=w<=0;
    if result then d:=a*b;
end;

{*************************************************************************}
{******* conditional assignment ******************************************}
{*************************************************************************}
//define[@select<][>][
function select(const cond:boolean; const vtrue,vfalse:#0):#0; inline;
begin
    if cond then result:=vtrue
    else result:=vfalse;
end;
]
@select<shortstring>
@select<ansistring>
@select<longint>
@select<double>
@select<single>
@select<byte>
//udefine[@select<]
